/* ----------------------------------------------------------------------
 * Copyright (C) 2013 ARM Limited. All rights reserved.
 *  
 * $Date:        13. September 2013
 * $Revision:    V1.01
 *  
 * Driver:       Driver_MCI0
 * Configured:   via RTE_Device.h configuration file 
 * Project:      MCI Driver for Energy Micro EFM32GGxxx
 * ---------------------------------------------------------------------- 
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 * 
 *   Configuration Setting                 Value
 *   ---------------------                 -----
 *   Connect to hardware via Driver_MCI# = 0
 * -------------------------------------------------------------------- */

/* History:
 *  Version 1.01
 *    Based on API V1.10 (namespace prefix ARM_ added)
 *  Version 1.00
 *    Initial release
 */

#include "cmsis_os.h"
#include "em_device.h"

#include "em_gpio.h"

#include "Driver_MCI.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

#define ARM_MCI_DRV_VERSION ARM_DRV_VERSION_MAJOR_MINOR(1,01)  /* driver version */

/* Driver Version */
static const ARM_DRV_VERSION DriverVersion = {
  ARM_MCI_API_VERSION,
  ARM_MCI_DRV_VERSION
};

/* Driver Capabilities */
static const ARM_MCI_CAPABILITIES DriverCapabilities = {
  0,        /* cd_state      */
  0,        /* cd_event      */
  0,        /* wp_state      */
  0,        /* busy_event    */
  0,        /* data_width_4  */
  0,        /* data_width_8  */
  0,        /* power_off     */
  0,        /* vdd_1v8       */
  0,        /* signaling_1v8 */
  0         /* reserved      */
};


/* MCI Information */
static struct {
  bool                   Power;         // Power flag
  bool                   SPI_Mode;      // SPI mode
  ARM_MCI_SignalEvent_t  cb_event;      // Event Callback
} MCI_Info;


/**
  \fn          ARM_DRV_VERSION GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRV_VERSION GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_MCI_CAPABILITIES GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_MCI_CAPABILITIES
*/
static ARM_MCI_CAPABILITIES GetCapabilities (void) {
  return DriverCapabilities;
}


/**
  \fn          ARM_MCI_STATUS MCI_Initialize (ARM_MCI_SignalEvent_t cb_event, bool spi_mode)
  \brief       Initialize the Memory Card Interface
  \param[in]   cb_event  Pointer to \ref ARM_MCI_SignalEvent
  \param[in]   spi_mode  SPI Mode (default is native MMC/SD Mode)
  \return      \ref ARM_MCI_STATUS execution status
*/
static ARM_MCI_STATUS Initialize (ARM_MCI_SignalEvent_t cb_event, bool spi_mode) {

  MCI_Info.cb_event = cb_event;

  /* Configure CD (Card Detect) Pin */

  /* Configure WP (Write Protect) Pin */

  MCI_Info.SPI_Mode = spi_mode;
  if (spi_mode == true) {
    return ARM_MCI_OK;
  }

  /* Native mode unsupported */
  return ARM_MCI_ERROR_UNSUPPORTED;
}


/**
  \fn          ARM_MCI_STATUS MCI_Uninitialize (void)
  \brief       De-initialize Memory Card Interface.
  \return      \ref ARM_MCI_STATUS execution status
*/
static ARM_MCI_STATUS Uninitialize (void) {
  MCI_Info.Power = false;

  /* Unconfigure CD (Card Detect) Pin */

  /* Unconfigure WP (Write Protect) Pin */

  if (MCI_Info.SPI_Mode == true) {
    return ARM_MCI_OK;
  }

  return ARM_MCI_ERROR_UNSUPPORTED;;
}


/**
  \fn          ARM_MCI_STATUS MCI_PowerControl (ARM_POWER_STATE state)
  \brief       Control Memory Card Interface Power.
  \param[in]   state   Power state \ref ARM_POWER_STATE
  \return      \ref ARM_MCI_STATUS execution status
*/
static ARM_MCI_STATUS PowerControl (ARM_POWER_STATE state) {

  switch (state) {
    case ARM_POWER_OFF:
      MCI_Info.Power = false;
      break;

    case ARM_POWER_LOW:
      return ARM_MCI_ERROR_UNSUPPORTED;
      
    case ARM_POWER_FULL:
      MCI_Info.Power = true;
      break;

    default:
      return ARM_MCI_ERROR_UNSUPPORTED;
  }
  return ARM_MCI_OK;
}


/**
  \fn          ARM_MCI_SWITCH MCI_CardSwitchRead (void)
  \brief       Read state of Memory Card switches.
  \return      \ref ARM_MCI_SWITCH execution status
*/
static ARM_MCI_SWITCH CardSwitchRead (void) {
  ARM_MCI_SWITCH sw = { 0, 0 };

  /* Read CD (Card Detect) Pin */
  sw.cd_state = 1;

  /* Read WP (Write Protect) Pin */
  sw.wp_state = 0;

  return (sw);
}



/* MCI Driver Control Block */
ARM_DRIVER_MCI Driver_MCI0 = {
  GetVersion,
  GetCapabilities,
  Initialize,
  Uninitialize,
  PowerControl,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  CardSwitchRead,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL
};
