/* ----------------------------------------------------------------------
 * Copyright (C) 2013 ARM Limited. All rights reserved.
 *  
 * $Date:        13. September 2013
 * $Revision:    V1.01
 *  
 * Project:      OTG Full/Low-Speed Common Driver
 *               for EnergyMicro EFM32GGxxx
 * Configured:   via RTE_Device.h configuration file 
 * -------------------------------------------------------------------- */

/* History:
 *  Version 1.01
 *    Based on API V1.10 (namespace prefix ARM_ added)
 *  Version 1.00
 *    Initial release
 */

#include <stdint.h>
#include <string.h>
#include "cmsis_os.h"

#include "Driver_USBH.h"
#include "Driver_USBD.h"

#include "em_device.h"

#include "em_cmu.h"
#include "em_emu.h"
#include "em_gpio.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

#if (RTE_USB_HOST)
extern void USBH_IRQ (uint32_t gintsts);
#endif
#if (RTE_USB_DEV)
extern void USBD_IRQ (uint32_t gintsts);
#endif

static bool    pins_configured_mask = 0;
       uint8_t otg_role             = ARM_USB_ROLE_NONE;


/* Common IRQ Routine *********************************************************/

/**
  \fn          void OTG_IRQHandler (void)
  \brief       USB Interrupt Routine (IRQ).
*/
void USB_IRQHandler (void) {
  uint32_t gintsts;

  gintsts = USB->GINTSTS & USB->GINTMSK;

  switch (otg_role) {
    case ARM_USB_ROLE_HOST:
#if (RTE_USB_HOST)
      USBH_IRQ (gintsts);
#endif
      break;
    case ARM_USB_ROLE_DEVICE:
#if (RTE_USB_DEV)
      USBD_IRQ (gintsts);
#endif
      break;
    case ARM_USB_ROLE_NONE:
      break;
  }
}


/* Auxiliary Functions ********************************************************/

/* Pin Functions ---------------------*/

/**
  \fn          bool OTG_PinsConfigure (uint8_t pins_mask)
  \brief       Configure single or multiple USB Pin(s).
  \param[in]   Mask of pins to be configured (possible masking values:
               USB_PIN_DP, USB_PIN_DM, USB_PIN_VBUS, USB_PIN_OC, USB_PIN_ID)
  \return      true = success, false = fail
*/
bool OTG_PinsConfigure (uint8_t pins_mask) {

#if (RTE_USB)
#if (RTE_USB_HOST && !RTE_USB_VBUS_PIN)
#error   Configure VBUS Power On/Off pin for USB OTG Full-speed in RTE_Device.h!
#endif
#else
#error   Enable USB Full-speed in RTE_Device.h!
#endif

  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_DP) {
    // Init PHY. Enable USB pins (DP and DM)
    CMU->HFCORECLKEN0    |= CMU_HFCORECLKEN0_USB;
    USB->ROUTE           |= USB_ROUTE_PHYPEN;
    pins_configured_mask |= ARM_USB_PIN_DP | ARM_USB_PIN_DM;
  }
  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_DM) {
    // Init PHY. Enable USB pins (DP and DM)
    CMU->HFCORECLKEN0    |= CMU_HFCORECLKEN0_USB;
    USB->ROUTE           |= USB_ROUTE_PHYPEN;
    pins_configured_mask |= ARM_USB_PIN_DM | ARM_USB_PIN_DP;
  }
  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_ID) {
    pins_configured_mask |= ARM_USB_PIN_ID;
  }
  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_VBUS) {
    // Enable VBUS pin
    CMU_ClockEnable( cmuClock_GPIO, true );
    GPIO_PinModeSet( RTE_USB_VBUS_PORT, RTE_USB_VBUS_BIT, gpioModePushPull, 0 );
    CMU->HFCORECLKEN0    |= CMU_HFCORECLKEN0_USB;
    USB->ROUTE           |= USB_ROUTE_VBUSENPEN;
    pins_configured_mask |= ARM_USB_PIN_VBUS;
  }
  if ((pins_configured_mask ^ pins_mask) & ARM_USB_PIN_OC) {
    GPIO_PinModeSet( RTE_USB_OC_PORT, RTE_USB_OC_BIT, gpioModeInput, 0 );
    pins_configured_mask |= ARM_USB_PIN_OC;
  }

  return true;
}

/**
  \fn          bool OTG_PinsUnconfigure (uint8_t pins_mask)
  \brief       Unconfigure to reset settings single or multiple USB Pin(s).
  \param[in]   Mask of pins to be unconfigured (possible masking values:
               USB_PIN_DP, USB_PIN_DM, USB_PIN_VBUS, USB_PIN_OC, USB_PIN_ID)
  \return      true = success, false = fail
*/
bool OTG_PinsUnconfigure (uint8_t pins_mask) {

  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_DP) {
    // Disable USB pins (DP and DM)
    USB->ROUTE           &= ~USB_ROUTE_PHYPEN;
    pins_configured_mask &= ~(ARM_USB_PIN_DP | ARM_USB_PIN_DM);
  }
  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_DM) {
    // Disable USB pins (DP and DM)
    USB->ROUTE           &= ~USB_ROUTE_PHYPEN;
    pins_configured_mask &= ~(ARM_USB_PIN_DM | ARM_USB_PIN_DP);
  }
  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_ID) {
    pins_configured_mask &= ~ARM_USB_PIN_ID;
  }
  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_VBUS) {
    // Disable VBUS pin
    USB->ROUTE &= ~USB_ROUTE_VBUSENPEN;
    GPIO_PinModeSet( RTE_USB_VBUS_PORT, RTE_USB_VBUS_BIT, gpioModeDisabled, 0 );
    pins_configured_mask &= ~ARM_USB_PIN_VBUS;
  }
  if ((pins_configured_mask & pins_mask) & ARM_USB_PIN_OC) {
    /* Enable VBUS overcurrent flag pin. */
    GPIO_PinModeSet( RTE_USB_OC_PORT, RTE_USB_OC_BIT, gpioModeDisabled, 0 );
    pins_configured_mask &= ~ARM_USB_PIN_OC;
  }

  return true;
}

/**
  \fn          bool OTG_PinVbusOnOff (bool state)
  \brief       Drive VBUS Pin On/Off.
  \param[in]   state    State On/Off (true = On, false = Off)
  \return      true = success, false = fail
*/

bool OTG_PinVbusOnOff (bool state) {
#if (RTE_USB_VBUS_PIN)
  return true;
#else
  return false;
#endif
}
