/* ----------------------------------------------------------------------
 * Copyright (C) 2013 ARM Limited. All rights reserved.
 *  
 * $Date:        09. October 2013
 * $Revision:    V1.02
 *  
 * Driver:       Driver_SPI0, Driver_SPI1, Driver_SPI2
 * Configured:   via RTE_Device.h configuration file 
 * Project:      SPI Driver for Energy Micro EFM32GGxxx
 * ---------------------------------------------------------------------- 
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 * 
 *   Configuration Setting               Value     SPI Interface
 *   ---------------------               -----     -------------
 *   Connect to hardware via Driver_SPI# = 0       use USART0
 *   Connect to hardware via Driver_SPI# = 1       use USART1
 *   Connect to hardware via Driver_SPI# = 2       use USART2
 * -------------------------------------------------------------------- */

/* History:
 *  Version 1.02
 *    Corrected DMA transfer mode for packet size > 1024 bytes
 *    Replaced usage of UART_ macros with correct USART_ variants.
 *  Version 1.01
 *    Based on API V1.10 (namespace prefix ARM_ added)
 *  Version 1.00
 *    Initial release
 */

#include "cmsis_os.h"
#include "em_device.h"

#include "em_cmu.h"
#include "em_dma.h"
#include "em_gpio.h"

#include "Driver_SPI.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

#define ARM_SPI_DRV_VERSION ARM_DRV_VERSION_MAJOR_MINOR(1,02) /* driver version */

#if ((defined(RTE_Drivers_SPI0) || \
      defined(RTE_Drivers_SPI1) || \
      defined(RTE_Drivers_SPI2))   \
     && !RTE_USART0                \
     && !RTE_USART1                \
     && !RTE_USART2)
#error "USART not configured in RTE_Device.h!"
#endif

#if ((RTE_USART0 && RTE_USART0_SYNC && RTE_USART0_TX_DMA) || \
     (RTE_USART1 && RTE_USART1_SYNC && RTE_USART1_TX_DMA) || \
     (RTE_USART2 && RTE_USART2_SYNC && RTE_USART2_TX_DMA))
  #define USART_TX_DMA_ENABLE
#endif

#if ((RTE_USART0 && RTE_USART0_SYNC && RTE_USART0_RX_DMA) || \
     (RTE_USART1 && RTE_USART1_SYNC && RTE_USART1_RX_DMA) || \
     (RTE_USART2 && RTE_USART2_SYNC && RTE_USART2_RX_DMA))
  #define USART_RX_DMA_ENABLE
#endif

/* USART configuration flags */
#define USART_CFG_CLK_ENABLE      (1U << 0)
#define USART_CFG_CS_ENABLE       (1U << 1)
#define USART_CFG_CS_MANUAL       (1U << 2)
#define USART_CFG_CS_ACTIVE_HIGH  (1U << 3)
#define USART_CFG_RX_DMA_ENABLE   (1U << 4)
#define USART_CFG_TX_DMA_ENABLE   (1U << 5)

#define DMA_CH_CTRL_SIGSEL_USARTRXDATAV (0)
#define DMA_CH_CTRL_SIGSEL_USARTTXBL    (1)

/* USART Transfer Timeout in microseconds */
#define USART_TRANSFER_TIMEOUT       500000


/* Driver Version */
static const ARM_DRV_VERSION DriverVersion = {
  ARM_SPI_API_VERSION,
  ARM_SPI_DRV_VERSION
};

/* Driver Capabilities */
static const ARM_SPI_CAPABILITIES DriverCapabilities = {
  1,  /* send_data_event */
  0
};


/* SPI Data Information (Run-Time) */
typedef struct {
  uint8_t              *buf;            // Pointer to data buffer
  uint32_t              cnt;            // Number of data bytes
  bool                  inc;            // Buffer increment flag
} SPI_DATA_INFO;

/* SPI Information (Run-Time) */
typedef struct {
  bool                  power;          // Power flag
  osSemaphoreId         semaphore_xfer; // Transfer Semaphore
  osSemaphoreId         semaphore_busy; // Busy Semaphore
  osSemaphoreId         semaphore_part; // Partial Transfer Semaphore
  ARM_SPI_SignalEvent_t cb_event;       // Event Callback
  SPI_DATA_INFO         rx;             // Receive information
  SPI_DATA_INFO         tx;             // Transmit information
  bool                  part;           // Partial transfer flag
} SPI_INFO;

/* GPIO pin description */
typedef struct _GPIO_PIN_ID {
  GPIO_Port_TypeDef     port;
  uint8_t               num;
} GPIO_PIN_ID;


/* USART Resources definition */
typedef struct {
        USART_TypeDef    *usart;                // Pointer to USART peripheral
  enum  IRQn              usart_rx_irq;         // USART Rx IRQ Number
  enum  IRQn              usart_tx_irq;         // USART Tx IRQ Number
        GPIO_PIN_ID       pin_rx;               // RX/MISO Pin identifier
        GPIO_PIN_ID       pin_tx;               // TX/MOSI Pin identifier
        GPIO_PIN_ID       pin_clk;              // CLK Pin identifier
        GPIO_PIN_ID       pin_cs;               // CS Pin identifier
        uint8_t           pin_loc;              // USART Pin Location
        uint8_t           dma_rx_ch;            // USART Receive DMA channel
        uint8_t           dma_tx_ch;            // USART Transmit DMA channel
        uint32_t          dma_sourcesel;        // DMA channel input source
        uint32_t          clock_limit;          // Maximum SPI bus frequency
        CMU_Clock_TypeDef clock_mask;           // Clock register mask
        uint8_t           cfg;                  // Configuration flags
  const osSemaphoreDef_t *semaphore_xfer;       // Transfer Semaphore definition
  const osSemaphoreDef_t *semaphore_busy;       // Busy Semaphore definition
  const osSemaphoreDef_t *semaphore_part;       // Partial transfer Semaphore definition
        SPI_INFO         *info;                 // Run-Time information
} const SPI_RESOURCES;

#if defined(USART_TX_DMA_ENABLE) || defined(USART_RX_DMA_ENABLE)
/* DMA Descriptor array */
extern DMA_DESCRIPTOR_TypeDef DMA_DescArray[];

/* DMA Callback function */
void USART_DMA_Complete (uint32_t channel, bool primary, void *ptr_spi);

/* DMA Callback */
DMA_CB_TypeDef USART_DMA_Cb = {
  USART_DMA_Complete,
  NULL,
  0
};
#endif

#if (RTE_USART0 && RTE_USART0_SYNC)

#ifndef __USART0_CLK_MAX_FREQUENCY
  /* This definition is used to limit maximum bus clock frequency */
  #define __USART0_CLK_MAX_FREQUENCY    0
#endif

/* SPI0 Semaphores */
osSemaphoreDef(SPI0_Semaphore_Xfer);
osSemaphoreDef(SPI0_Semaphore_Busy);
osSemaphoreDef(SPI0_Semaphore_Part);
osSemaphoreId  SPI0_Semaphore_Xfer;
osSemaphoreId  SPI0_Semaphore_Busy;

/* SPI0 Information (Run-Time) */
static SPI_INFO SPI0_Info;

static SPI_RESOURCES SPI0_Resources = {
  USART0,
  USART0_RX_IRQn,
  USART0_TX_IRQn,
  { RTE_USART0_RX_PORT,     RTE_USART0_RX_BIT     },
  { RTE_USART0_TX_PORT,     RTE_USART0_TX_BIT     },
  { RTE_USART0_CLK_PORT,    RTE_USART0_CLK_BIT    },
  #if (RTE_USART0_CS_MAN)
  { RTE_USART0_CS_MAN_PORT, RTE_USART0_CS_MAN_BIT },
  #else
  { RTE_USART0_CS_PORT,     RTE_USART0_CS_BIT     },
  #endif
  RTE_USART0_PIN_LOC,
  RTE_USART0_RX_DMA_CHANNEL,
  RTE_USART0_TX_DMA_CHANNEL,
  DMA_CH_CTRL_SOURCESEL_USART0,
  __USART0_CLK_MAX_FREQUENCY,
  cmuClock_USART0,
  #if (RTE_USART0_CLK_PIN)
  USART_CFG_CLK_ENABLE     |
  #endif
  #if (RTE_USART0_CS_PIN)
  USART_CFG_CS_ENABLE      |
  #endif
  #if (RTE_USART0_CS_PIN && RTE_USART0_CS_MAN)
  USART_CFG_CS_MANUAL      |
  #endif
  #if (RTE_USART0_CS_PIN && RTE_USART0_CS_ACTIVE)
  USART_CFG_CS_ACTIVE_HIGH |
  #endif
  #if (RTE_USART0_RX_DMA)
  USART_CFG_RX_DMA_ENABLE  |
  #endif
  #if (RTE_USART0_TX_DMA)
  USART_CFG_TX_DMA_ENABLE  |
  #endif
  0,
  osSemaphore(SPI0_Semaphore_Xfer),
  osSemaphore(SPI0_Semaphore_Busy),
  osSemaphore(SPI0_Semaphore_Part),
  &SPI0_Info
};
#endif /* RTE_USART0 && RTE_USART0_SYNC */

#if (RTE_USART1 && RTE_USART1_SYNC)

#ifndef __USART1_CLK_MAX_FREQUENCY
  /* This definition is used to limit maximum bus clock frequency */
  #define __USART1_CLK_MAX_FREQUENCY    0
#endif

/* SPI1 Semaphores */
osSemaphoreDef(SPI1_Semaphore_Xfer);
osSemaphoreDef(SPI1_Semaphore_Busy);
osSemaphoreDef(SPI1_Semaphore_Part);
osSemaphoreId  SPI1_Semaphore_Xfer;
osSemaphoreId  SPI1_Semaphore_Busy;

/* SPI0 Information (Run-Time) */
static SPI_INFO SPI1_Info;

static SPI_RESOURCES SPI1_Resources = {
  USART1,
  USART1_RX_IRQn,
  USART1_TX_IRQn,
  { RTE_USART1_RX_PORT,     RTE_USART1_RX_BIT     },
  { RTE_USART1_TX_PORT,     RTE_USART1_TX_BIT     },
  { RTE_USART1_CLK_PORT,    RTE_USART1_CLK_BIT    },
  #if (RTE_USART1_CS_MAN)
  { RTE_USART1_CS_MAN_PORT, RTE_USART1_CS_MAN_BIT },
  #else
  { RTE_USART1_CS_PORT,     RTE_USART1_CS_BIT     },
  #endif
  RTE_USART1_PIN_LOC,
  RTE_USART1_RX_DMA_CHANNEL,
  RTE_USART1_TX_DMA_CHANNEL,
  DMA_CH_CTRL_SOURCESEL_USART1,
  __USART1_CLK_MAX_FREQUENCY,
  cmuClock_USART1,
  #if (RTE_USART1_CLK_PIN)
  USART_CFG_CLK_ENABLE     |
  #endif
  #if (RTE_USART1_CS_PIN)
  USART_CFG_CS_ENABLE      |
  #endif
  #if (RTE_USART1_CS_PIN && RTE_USART1_CS_MAN)
  USART_CFG_CS_MANUAL      |
  #endif
  #if (RTE_USART1_CS_PIN && RTE_USART1_CS_ACTIVE)
  USART_CFG_CS_ACTIVE_HIGH |
  #endif
  #if (RTE_USART1_RX_DMA)
  USART_CFG_RX_DMA_ENABLE  |
  #endif
  #if (RTE_USART1_TX_DMA)
  USART_CFG_TX_DMA_ENABLE  |
  #endif
  0,
  osSemaphore(SPI1_Semaphore_Xfer),
  osSemaphore(SPI1_Semaphore_Busy),
  osSemaphore(SPI1_Semaphore_Part),
  &SPI1_Info
};
#endif /* RTE_USART1 && RTE_USART1_SYNC */

#if (RTE_USART2 && RTE_USART2_SYNC)

#ifndef __USART2_CLK_MAX_FREQUENCY
  /* This definition is used to limit maximum bus clock frequency */
  #define __USART2_CLK_MAX_FREQUENCY    0
#endif

/* SPI2 Semaphores */
osSemaphoreDef(SPI2_Semaphore_Xfer);
osSemaphoreDef(SPI2_Semaphore_Busy);
osSemaphoreDef(SPI2_Semaphore_Part);
osSemaphoreId  SPI2_Semaphore_Xfer;
osSemaphoreId  SPI2_Semaphore_Busy;

/* SPI2 Information (Run-Time) */
static SPI_INFO SPI2_Info;

static SPI_RESOURCES SPI2_Resources = {
  USART2,
  USART2_RX_IRQn,
  USART2_TX_IRQn,
  { RTE_USART2_RX_PORT,     RTE_USART2_RX_BIT     },
  { RTE_USART2_TX_PORT,     RTE_USART2_TX_BIT     },
  { RTE_USART2_CLK_PORT,    RTE_USART2_CLK_BIT    },
  #if (RTE_USART2_CS_MAN)
  { RTE_USART2_CS_MAN_PORT, RTE_USART2_CS_MAN_BIT },
  #else
  { RTE_USART2_CS_PORT,     RTE_USART2_CS_BIT     },
  #endif
  RTE_USART2_PIN_LOC,
  RTE_USART2_RX_DMA_CHANNEL,
  RTE_USART2_TX_DMA_CHANNEL,
  DMA_CH_CTRL_SOURCESEL_USART2,
  __USART2_CLK_MAX_FREQUENCY,
  cmuClock_USART2,
  #if (RTE_USART2_CLK_PIN)
  USART_CFG_CLK_ENABLE     |
  #endif
  #if (RTE_USART2_CS_PIN)
  USART_CFG_CS_ENABLE      |
  #endif
  #if (RTE_USART2_CS_PIN && RTE_USART2_CS_MAN)
  USART_CFG_CS_MANUAL      |
  #endif
  #if (RTE_USART2_CS_PIN && RTE_USART2_CS_ACTIVE)
  USART_CFG_CS_ACTIVE_HIGH |
  #endif
  #if (RTE_USART2_RX_DMA)
  USART_CFG_RX_DMA_ENABLE  |
  #endif
  #if (RTE_USART2_TX_DMA)
  USART_CFG_TX_DMA_ENABLE  |
  #endif
  0,
  osSemaphore(SPI2_Semaphore_Xfer),
  osSemaphore(SPI2_Semaphore_Busy),
  osSemaphore(SPI2_Semaphore_Part),
  &SPI2_Info
};
#endif /* RTE_USART2 && RTE_USART2_SYNC */

/**
  \fn          ARM_DRV_VERSION SPI_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRV_VERSION SPIX_GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_SPI_CAPABILITIES SPI_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_SPI_CAPABILITIES
*/
static ARM_SPI_CAPABILITIES SPIX_GetCapabilities (void) {
  return DriverCapabilities;
}


/**
  \fn          ARM_SPI_STATUS SPI_Initialize (ARM_SPI_SignalEvent_t cb_event,
                                              SPI_RESOURCES        *ptr_spi)
  \brief       Initialize SPI Interface.
  \param[in]   cb_event  Pointer to \ref ARM_SPI_SignalEvent
  \param[in]   ptr_spi   Pointer to SPI resources
  \return      \ref ARM_SPI_STATUS
*/
static ARM_SPI_STATUS SPI_Initialize (ARM_SPI_SignalEvent_t cb_event,
                                      SPI_RESOURCES        *ptr_spi) {
  #if defined(USART_RX_DMA_ENABLE) || defined(USART_TX_DMA_ENABLE)
  DMA_Init_TypeDef DmaInit;
  #endif

  /* Enable USARTx peripheral clock */
  CMU_ClockEnable (ptr_spi->clock_mask, true);

  /* Enable GPIO peripheral clock */
  CMU_ClockEnable (cmuClock_GPIO, true);

  /* Configure USART TX pin (MOSI) */
  GPIO_PinModeSet (ptr_spi->pin_tx.port, ptr_spi->pin_tx.num, gpioModePushPull, 0);

  /* Configure USART RX pin (MISO) */
  GPIO_PinModeSet (ptr_spi->pin_rx.port, ptr_spi->pin_rx.num, gpioModePushPull, 0);

  if (ptr_spi->cfg & USART_CFG_CLK_ENABLE) {
    /* Configure USART CLK pin (SCLK) */
    GPIO_PinModeSet (ptr_spi->pin_clk.port, ptr_spi->pin_clk.num, gpioModePushPull, 0);
  }

  if (ptr_spi->cfg & USART_CFG_CS_ENABLE) {
    /* Configure USART CS pin (SS) */
    GPIO_PinModeSet (ptr_spi->pin_cs.port, ptr_spi->pin_cs.num, gpioModePushPull, 0);
  }

  /* Define USARTx pin location and enable Tx, Rx, CLK and CS pins */
  ptr_spi->usart->ROUTE = (ptr_spi->usart->ROUTE & ~_USART_ROUTE_LOCATION_MASK)  |
                          (ptr_spi->pin_loc      << _USART_ROUTE_LOCATION_SHIFT) |
                           USART_ROUTE_RXPEN | USART_ROUTE_TXPEN;

  if (ptr_spi->cfg & USART_CFG_CLK_ENABLE) {
    /* Enable CLK pin */
    ptr_spi->usart->ROUTE |= USART_ROUTE_CLKPEN;
  }

  if ((ptr_spi->cfg & USART_CFG_CS_ENABLE) && !(ptr_spi->cfg & USART_CFG_CS_MANUAL)) {
    /* CS pin in auto mode */
    ptr_spi->usart->ROUTE |= USART_ROUTE_CSPEN;
  }

  /* Initialize DMA */
  #if defined (USART_TX_DMA_ENABLE) || defined (USART_RX_DMA_ENABLE)
  if (ptr_spi->cfg & (USART_CFG_RX_DMA_ENABLE | USART_CFG_TX_DMA_ENABLE)) {
    DmaInit.hprot        = false;
    DmaInit.controlBlock = DMA_DescArray;
    DMA_Init (&DmaInit);
  }
  #endif

  /* Initialize Run-Time Resources */
  ptr_spi->info->power = false;
  ptr_spi->info->semaphore_xfer = osSemaphoreCreate(ptr_spi->semaphore_xfer, 0);
  ptr_spi->info->semaphore_busy = osSemaphoreCreate(ptr_spi->semaphore_busy, 0);
  ptr_spi->info->semaphore_part = osSemaphoreCreate(ptr_spi->semaphore_part, 0);
  if (ptr_spi->info->semaphore_xfer == NULL) return ARM_SPI_ERROR;
  if (ptr_spi->info->semaphore_busy == NULL) return ARM_SPI_ERROR;
  if (ptr_spi->info->semaphore_part == NULL) return ARM_SPI_ERROR;

  ptr_spi->info->cb_event = cb_event;

  /* Enable USART interrupts */
  NVIC_ClearPendingIRQ (ptr_spi->usart_tx_irq);
  NVIC_ClearPendingIRQ (ptr_spi->usart_rx_irq);
  NVIC_EnableIRQ (ptr_spi->usart_tx_irq);
  NVIC_EnableIRQ (ptr_spi->usart_rx_irq);

  return ARM_SPI_OK;
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static ARM_SPI_STATUS SPI0_Initialize (ARM_SPI_SignalEvent_t pSignalEvent) {
  return SPI_Initialize(pSignalEvent, &SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static ARM_SPI_STATUS SPI1_Initialize (ARM_SPI_SignalEvent_t pSignalEvent) {
  return SPI_Initialize(pSignalEvent, &SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
static ARM_SPI_STATUS SPI2_Initialize (ARM_SPI_SignalEvent_t pSignalEvent) {
  return SPI_Initialize(pSignalEvent, &SPI2_Resources);
}
#endif


/**
  \fn          ARM_SPI_STATUS SPI_Uninitialize (SPI_RESOURCES *ptr_spi)
  \brief       De-initialize SPI Interface.
  \param[in]   ptr_spi  Pointer to SPI resources
  \return      \ref ARM_SPI_STATUS
*/
static ARM_SPI_STATUS SPI_Uninitialize (SPI_RESOURCES *ptr_spi) {

  /* Enable USART peripheral clock */
  CMU_ClockEnable (ptr_spi->clock_mask, true);

  /* Disable USART IRQ */
  NVIC_DisableIRQ (ptr_spi->usart_tx_irq);
  NVIC_DisableIRQ (ptr_spi->usart_rx_irq);

  /* Disable USART pins */
  ptr_spi->usart->ROUTE = 0;

  /* Disable USART peripheral clock */
  CMU_ClockEnable (ptr_spi->clock_mask, false);

  /* Unconfigure USART Tx pin */
  GPIO_PinModeSet (ptr_spi->pin_tx.port, ptr_spi->pin_tx.num, gpioModeDisabled, 0);
  /* Unconfigure USART Rx pin */
  GPIO_PinModeSet (ptr_spi->pin_rx.port, ptr_spi->pin_rx.num, gpioModeDisabled, 0);

  if (ptr_spi->pin_clk.port) {
    /* Unconfigure USART CLK pin */
    GPIO_PinModeSet (ptr_spi->pin_clk.port, ptr_spi->pin_clk.num, gpioModeDisabled, 0);
  }

  if (ptr_spi->pin_cs.port) {
    /* Unconfigure USART CS pin */
    GPIO_PinModeSet (ptr_spi->pin_cs.port, ptr_spi->pin_cs.num, gpioModeDisabled, 0);
  }

  return ARM_SPI_OK;
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static ARM_SPI_STATUS SPI0_Uninitialize (void) {
  return SPI_Uninitialize (&SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static ARM_SPI_STATUS SPI1_Uninitialize (void) {
  return SPI_Uninitialize (&SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
static ARM_SPI_STATUS SPI2_Uninitialize (void) {
  return SPI_Uninitialize (&SPI2_Resources);
}
#endif


/**
  \fn          ARM_SPI_STATUS SPI_PowerControl (ARM_POWER_STATE state,
                                                SPI_RESOURCES  *ptr_spi)
  \brief       Controls SPI Interface Power.
  \param[in]   state     Power state
  \param[in]   ptr_spi   Pointer to SPI resources
  \return      \ref ARM_SPI_STATUS
*/
static ARM_SPI_STATUS SPI_PowerControl (ARM_POWER_STATE state, SPI_RESOURCES *ptr_spi) {
  #if defined(USART_RX_DMA_ENABLE) || defined(USART_TX_DMA_ENABLE)
  DMA_CfgChannel_TypeDef DmaCfg;
  #endif
  uint32_t ctrl;

  switch (state) {
    case ARM_POWER_OFF:
      ptr_spi->info->power = false;

      /* Disable USART clock */
      CMU_ClockEnable (ptr_spi->clock_mask, false);
      break;

    case ARM_POWER_LOW:
      break;

    case ARM_POWER_FULL:
      /* Enable USARTx peripheral clock */
      CMU_ClockEnable (ptr_spi->clock_mask, true);
      
      #if defined(USART_RX_DMA_ENABLE) || defined(USART_TX_DMA_ENABLE)
      /* Initialize DMA channels */
      if (ptr_spi->cfg & (USART_CFG_RX_DMA_ENABLE | USART_CFG_TX_DMA_ENABLE)) {
        /* Setup DMA channel configuration structure */
        DmaCfg.cb        = &USART_DMA_Cb;
        DmaCfg.enableInt = true;
        DmaCfg.highPri   = false;

        if (ptr_spi->cfg & USART_CFG_RX_DMA_ENABLE) {
          DmaCfg.select = ptr_spi->dma_sourcesel | DMA_CH_CTRL_SIGSEL_USARTRXDATAV;
          /* Configure channel */
          DMA_CfgChannel (ptr_spi->dma_rx_ch, &DmaCfg);
        }

        if (ptr_spi->cfg & USART_CFG_TX_DMA_ENABLE) {
          DmaCfg.select = ptr_spi->dma_sourcesel | DMA_CH_CTRL_SIGSEL_USARTTXBL;
          /* Configure channel */
          DMA_CfgChannel (ptr_spi->dma_tx_ch, &DmaCfg);
        }
      }
      #endif

      /* Clear buffers, shift registers and enable master mode */
      ptr_spi->usart->CMD = USART_CMD_CLEARRX | USART_CMD_CLEARTX | USART_CMD_MASTEREN;
      
      /* Enable synchronous mode and set USART control register */
      ctrl = USART_CTRL_SYNC;
      if (!(ptr_spi->cfg & USART_CFG_CS_MANUAL)) {
        ctrl |= USART_CTRL_AUTOCS;
        if (ptr_spi->cfg & USART_CFG_CS_ACTIVE_HIGH) {
          ctrl |= USART_CTRL_CSINV;
        }
      }
      ptr_spi->usart->CTRL |= ctrl;

      /* Enable Rx and Tx */
      ptr_spi->usart->CMD = USART_CMD_RXEN | USART_CMD_TXEN;
      
      /* Ready for operation */
      osSemaphoreRelease(ptr_spi->info->semaphore_busy);
      ptr_spi->info->power = true;
      break;
  }
  return ARM_SPI_OK;
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static ARM_SPI_STATUS SPI0_PowerControl (ARM_POWER_STATE state) {
  return SPI_PowerControl(state, &SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static ARM_SPI_STATUS SPI1_PowerControl (ARM_POWER_STATE state) {
  return SPI_PowerControl(state, &SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
static ARM_SPI_STATUS SPI2_PowerControl (ARM_POWER_STATE state) {
  return SPI_PowerControl(state, &SPI2_Resources);
}
#endif


/**
  \fn          ARM_SPI_STATUS SPI_Configure (ARM_SPI_FRAME_FORMAT frame_format,
                                             ARM_SPI_BIT_ORDER    bit_order,
                                             SPI_RESOURCES       *ptr_spi)
  \brief       Configure SPI Interface.
  \param[in]   frame_format  \ref ARM_SPI_FRAME_FORMAT
  \param[in]   bit_order     \ref ARM_SPI_BIT_ORDER
  \param[in]   ptr_spi       Pointer to SPI resources
  \return      \ref SPI_STATUS
*/
static ARM_SPI_STATUS SPI_Configure (ARM_SPI_FRAME_FORMAT frame_format,
                                     ARM_SPI_BIT_ORDER    bit_order,
                                     SPI_RESOURCES       *ptr_spi) {
  uint32_t ctrl;
  
  if  (!ptr_spi->info->power) return ARM_SPI_ERROR;

  switch (frame_format) {
    case ARM_SPI_CPOL0_CPHA0:
      ctrl = 0;
      break;
    case ARM_SPI_CPOL0_CPHA1:
      ctrl = USART_CTRL_CLKPHA;
      break;
    case ARM_SPI_CPOL1_CPHA0:
      ctrl = USART_CTRL_CLKPOL;
      break;
    case ARM_SPI_CPOL1_CPHA1:
      ctrl = USART_CTRL_CLKPHA | USART_CTRL_CLKPOL;
      break;
    default:
      return ARM_SPI_ERROR_UNSUPPORTED;
  }

  switch (bit_order) {
    case ARM_SPI_MSB_LSB:
      ctrl |= USART_CTRL_MSBF;
      break;
    case ARM_SPI_LSB_MSB:
      break;
    default:
      return ARM_SPI_ERROR_UNSUPPORTED;
  }

  ptr_spi->usart->CTRL &= ~(USART_CTRL_MSBF | USART_CTRL_CLKPHA | USART_CTRL_CLKPOL);
  ptr_spi->usart->CTRL |= ctrl;
  return ARM_SPI_OK;
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static ARM_SPI_STATUS SPI0_Configure (ARM_SPI_FRAME_FORMAT  frame_format,
                                      ARM_SPI_BIT_ORDER     bit_order) {
  return SPI_Configure(frame_format, bit_order, &SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static ARM_SPI_STATUS SPI1_Configure (ARM_SPI_FRAME_FORMAT  frame_format,
                                      ARM_SPI_BIT_ORDER     bit_order) {
  return SPI_Configure(frame_format, bit_order, &SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
static ARM_SPI_STATUS SPI2_Configure (ARM_SPI_FRAME_FORMAT  frame_format,
                                      ARM_SPI_BIT_ORDER     bit_order) {
  return SPI_Configure(frame_format, bit_order, &SPI2_Resources);
}
#endif


/**
  \fn          uint32_t SPI_BusSpeed (uint32_t bps, SPI_RESOURCES *ptr_spi)
  \brief       Set bus speed for SPI transfers.
  \param[in]   bps      Requested speed in bits/s
  \param[in]   ptr_spi  Pointer to SPI resources
  \return      Configured bus speed in bits/s
*/
static uint32_t SPI_BusSpeed (uint32_t bps, SPI_RESOURCES *ptr_spi) {
  uint32_t div, per_clk;
  
  if (!ptr_spi->info->power) return 0;
  
  if (ptr_spi->clock_limit && bps > ptr_spi->clock_limit) {
    /* Limit maximum clock bitrate */
    bps = ptr_spi->clock_limit;
  }
  
  /* Check if USART operates in synchronous mode */
  if (ptr_spi->usart->CTRL & USART_CTRL_SYNC) {
    /* Get USART peripheral clock */
    per_clk = CMU_ClockFreqGet (ptr_spi->clock_mask);

    div  = (2 * per_clk + bps - 1) / bps;
    div -= 4;
    div *= 64;

    /* Round up */
    div += 0xc0;
    div &= 0xffffff00;

    ptr_spi->usart->CLKDIV = div & _USART_CLKDIV_DIV_MASK;
    return per_clk / (2 * (1 + div/256));
  }
  return (0);
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static uint32_t SPI0_BusSpeed (uint32_t bps) {
  return SPI_BusSpeed(bps, &SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static uint32_t SPI1_BusSpeed (uint32_t bps) {
  return SPI_BusSpeed(bps, &SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
static uint32_t SPI2_BusSpeed (uint32_t bps) {
  return SPI_BusSpeed(bps, &SPI2_Resources);
}
#endif


/**
  \fn          ARM_SPI_STATUS SPI_SlaveSelect (ARM_SPI_SS_SIGNAL ss,
                                               SPI_RESOURCES    *ptr_spi)
  \brief       Control Slave Select (SS) signal.
  \param[in]   ss       \ref ARM_SPI_SS_SIGNAL
  \param[in]   ptr_spi  Pointer to SPI resources
  \return      \ref ARM_SPI_STATUS
*/
static ARM_SPI_STATUS SPI_SlaveSelect (ARM_SPI_SS_SIGNAL ss,
                                       SPI_RESOURCES    *ptr_spi) {
  uint32_t val;

  if (!ptr_spi->info->power) return ARM_SPI_ERROR;

  if (ptr_spi->pin_cs.port) {
    val = (ptr_spi->cfg & USART_CFG_CS_ACTIVE_HIGH) ? (1) : (0);
    val = (ss == ARM_SPI_SS_ACTIVE) ? val : !val;
    
    if (val) {
      GPIO_PinOutSet (ptr_spi->pin_cs.port, ptr_spi->pin_cs.num);
    }
    else {
      GPIO_PinOutClear (ptr_spi->pin_cs.port, ptr_spi->pin_cs.num);
    }
  }

  return ARM_SPI_OK;
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static ARM_SPI_STATUS SPI0_SlaveSelect (ARM_SPI_SS_SIGNAL ss) {
  return SPI_SlaveSelect(ss, &SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static ARM_SPI_STATUS SPI1_SlaveSelect (ARM_SPI_SS_SIGNAL ss) {
  return SPI_SlaveSelect(ss, &SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
static ARM_SPI_STATUS SPI2_SlaveSelect (ARM_SPI_SS_SIGNAL ss) {
  return SPI_SlaveSelect(ss, &SPI2_Resources);
}
#endif


/**
  \fn          uint8_t SPI_TransferByte (uint8_t out, SPI_RESOURCES *ptr_spi)
  \brief       Send and receive one byte via SPI Interface.
  \param[in]   out      Byte to be sent to the SPI data output
  \param[in]   ptr_spi  Pointer to SPI resources
  \return      Byte received from the SPI data input
*/
static uint8_t SPI_TransferByte (uint8_t out, SPI_RESOURCES *ptr_spi) {
  uint32_t tick;
  bool     tx;

  if (!ptr_spi->info->power) return 0;

  ptr_spi->usart->CMD = USART_CMD_RXEN | USART_CMD_TXEN;
  
  tx   = true;
  tick = osKernelSysTick();
  do {
    if (tx && (ptr_spi->usart->STATUS & USART_STATUS_TXBL)) {
      /* TX Buffer empty */
      ptr_spi->usart->TXDATA = out;
      tx = false;
    }
    
    if (!tx && (ptr_spi->usart->STATUS & USART_STATUS_TXC)) {
      /* TX Complete */
      return (ptr_spi->usart->RXDATA);
    }
  }
  while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(USART_TRANSFER_TIMEOUT));
  /* Timeout */
  return (0xFF);
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static uint8_t SPI0_TransferByte (uint8_t out) {
  return SPI_TransferByte(out, &SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static uint8_t SPI1_TransferByte (uint8_t out) {
  return SPI_TransferByte(out, &SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
static uint8_t SPI2_TransferByte (uint8_t out) {
  return SPI_TransferByte(out, &SPI2_Resources);
}
#endif


/**
  \fn          ARM_SPI_STATUS SPI_SendData (const uint8_t  *buf,
                                                  uint32_t  len,
                                            SPI_RESOURCES  *ptr_spi)
  \brief       Send data block to SPI Slave.
  \param[in]   buf      Pointer to the buffer containing data to send
  \param[in]   len      Data buffer length in bytes
  \param[in]   ptr_spi  Pointer to SPI resources
  \return      \ref ARM_SPI_STATUS
*/
static ARM_SPI_STATUS SPI_SendData (const uint8_t  *buf,
                                          uint32_t  len,
                                    SPI_RESOURCES  *ptr_spi) {
  DMA_CfgDescr_TypeDef DmaDescCfg;
  uint8_t *pb;
  uint8_t  ch;
  uint32_t sz;

  if ((buf == NULL) || (len == 0)) return ARM_SPI_ERROR;

  osSemaphoreWait(ptr_spi->info->semaphore_busy, osWaitForever);

  if (!ptr_spi->info->power) return ARM_SPI_ERROR;
  
  ptr_spi->info->part = false;
  
  while (len) {
    pb = (uint8_t *)buf;
    sz = len;
    if (sz > 1024) {
      sz = 1024;
      ptr_spi->info->part = true;
    }
    else {
      ptr_spi->info->part = false;
    }
    buf += sz;
    len -= sz;
    
    if (ptr_spi->cfg & USART_CFG_RX_DMA_ENABLE) {
      ptr_spi->usart->CMD = USART_CMD_RXDIS;

      ptr_spi->info->rx.buf = NULL;

      DmaDescCfg.dstInc  = dmaDataIncNone;
      DmaDescCfg.srcInc  = dmaDataIncNone;
      DmaDescCfg.size    = dmaDataSize1;
      DmaDescCfg.arbRate = dmaArbitrate1;
      DmaDescCfg.hprot   = 0;

      DMA_CfgDescr (ptr_spi->dma_rx_ch, true, &DmaDescCfg);

      DMA_ActivateBasic (ptr_spi->dma_rx_ch,
                         true,
                         false,
                         (void *)&ch,
                         (void *)&ptr_spi->usart->RXDATA,
                         sz-1);
      ptr_spi->usart->CMD = USART_CMD_RXEN;
    }
    else {
      ptr_spi->info->rx.buf = NULL;
      ptr_spi->info->rx.cnt = sz;
      ptr_spi->info->rx.inc = false;

      /* Enable Rx Data Valid interrupt  */
      ptr_spi->usart->IEN |= USART_IEN_RXDATAV;
    }

    if (ptr_spi->cfg & USART_CFG_TX_DMA_ENABLE) {
      ptr_spi->usart->CMD = USART_CMD_TXDIS;
      
      DmaDescCfg.dstInc  = dmaDataIncNone;
      DmaDescCfg.srcInc  = dmaDataInc1;
      DmaDescCfg.size    = dmaDataSize1;
      DmaDescCfg.arbRate = dmaArbitrate1;
      DmaDescCfg.hprot   = 0;

      DMA_CfgDescr (ptr_spi->dma_tx_ch, true, &DmaDescCfg);

      DMA_ActivateBasic (ptr_spi->dma_tx_ch,
                         true,
                         false,
                         (void *)&ptr_spi->usart->TXDATA,
                         (void *)pb,
                         sz-1);

      ptr_spi->usart->CMD = USART_CMD_TXEN;
    }
    else {
      ptr_spi->info->tx.buf = pb;
      ptr_spi->info->tx.cnt = sz;
      ptr_spi->info->tx.inc = true;

      /* Enable interrupt on TX buffer level */
      ptr_spi->usart->IEN |= USART_IEN_TXBL;
    }

    if (ptr_spi->info->part) {
      osSemaphoreWait (ptr_spi->info->semaphore_part, osWaitForever);
    }
  }
  osSemaphoreWait(ptr_spi->info->semaphore_xfer, osWaitForever);

  return ARM_SPI_OK;
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static ARM_SPI_STATUS SPI0_SendData (const uint8_t *buf, uint32_t len) {
  return SPI_SendData(buf, len, &SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static ARM_SPI_STATUS SPI1_SendData (const uint8_t *buf, uint32_t len) {
  return SPI_SendData(buf, len, &SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART1_SYNC)
static ARM_SPI_STATUS SPI2_SendData (const uint8_t *buf, uint32_t len) {
  return SPI_SendData(buf, len, &SPI2_Resources);
}
#endif


/**
  \fn          ARM_SPI_STATUS SPI_ReceiveData (uint8_t       *buf,
                                               uint32_t       len,
                                               uint8_t        out,
                                               SPI_RESOURCES *ptr_spi)
  \brief       Receive data block from SPI Slave.
  \param[out]  buf      Pointer to the buffer for receiving data
  \param[in]   len      Data buffer length in bytes
  \param[in]   out      Byte to be sent repeatedly during receive
  \param[in]   ptr_spi  Pointer to SPI resources
  \return      \ref ARM_SPI_STATUS
*/
static ARM_SPI_STATUS SPI_ReceiveData (uint8_t       *buf,
                                       uint32_t       len,
                                       uint8_t        out,
                                       SPI_RESOURCES *ptr_spi) {
  DMA_CfgDescr_TypeDef DmaDescCfg;
  uint8_t *pb;
  uint32_t sz;

  if ((buf == NULL) || (len == 0)) return ARM_SPI_ERROR;

  osSemaphoreWait(ptr_spi->info->semaphore_busy, osWaitForever);

  if (!ptr_spi->info->power) return ARM_SPI_ERROR;
  
  ptr_spi->info->part = false;
  
  while (len) {
    pb = buf;
    sz = len;
    if (sz > 1024) {
      sz = 1024;
      ptr_spi->info->part = true;
    }
    else {
      ptr_spi->info->part = false;
    }
    buf += sz;
    len -= sz;

    if (ptr_spi->cfg & USART_CFG_RX_DMA_ENABLE) {
      ptr_spi->usart->CMD = USART_CMD_RXDIS;
    
      ptr_spi->info->rx.buf = pb;

      DmaDescCfg.dstInc  = dmaDataInc1;
      DmaDescCfg.srcInc  = dmaDataIncNone;
      DmaDescCfg.size    = dmaDataSize1;
      DmaDescCfg.arbRate = dmaArbitrate1;
      DmaDescCfg.hprot   = 0;
    
      DMA_CfgDescr (ptr_spi->dma_rx_ch, true, &DmaDescCfg);

      DMA_ActivateBasic (ptr_spi->dma_rx_ch,
                         true,
                         false,
                         (void *)pb,
                         (void *)&ptr_spi->usart->RXDATA,
                         sz-1);

      ptr_spi->usart->CMD = USART_CMD_RXEN;
    }
    else {
      ptr_spi->info->rx.buf = pb;
      ptr_spi->info->rx.cnt = sz;
      ptr_spi->info->rx.inc = true;

      /* Enable Rx Data Valid interrupt  */
      ptr_spi->usart->IEN |= USART_IEN_RXDATAV;
    }

    if (ptr_spi->cfg & USART_CFG_TX_DMA_ENABLE) {
      ptr_spi->usart->CMD = USART_CMD_TXDIS;

      DmaDescCfg.dstInc  = dmaDataIncNone;
      DmaDescCfg.srcInc  = dmaDataIncNone;
      DmaDescCfg.size    = dmaDataSize1;
      DmaDescCfg.arbRate = dmaArbitrate1;
      DmaDescCfg.hprot   = 0;

      DMA_CfgDescr (ptr_spi->dma_tx_ch, true, &DmaDescCfg);

      DMA_ActivateBasic (ptr_spi->dma_tx_ch,
                         true,
                         false,
                         (void *)&ptr_spi->usart->TXDATA,
                         (void *)&out,
                         sz-1);

      ptr_spi->usart->CMD = USART_CMD_TXEN;
    }
    else {
      ptr_spi->info->tx.buf = &out;
      ptr_spi->info->tx.cnt = sz;
      ptr_spi->info->tx.inc = false;

      /* Enable interrupt on TX buffer level */
      ptr_spi->usart->IEN |= USART_IEN_TXBL;
    }
    
    if (ptr_spi->info->part) {
      osSemaphoreWait (ptr_spi->info->semaphore_part, osWaitForever);
    }
  }
  osSemaphoreWait(ptr_spi->info->semaphore_xfer, osWaitForever);
  
  return ARM_SPI_OK;
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static ARM_SPI_STATUS SPI0_ReceiveData (uint8_t *buf, uint32_t len, uint8_t out) {
  return SPI_ReceiveData(buf, len, out, &SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static ARM_SPI_STATUS SPI1_ReceiveData (uint8_t *buf, uint32_t len, uint8_t out) {
  return SPI_ReceiveData(buf, len, out, &SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
static ARM_SPI_STATUS SPI2_ReceiveData (uint8_t *buf, uint32_t len, uint8_t out) {
  return SPI_ReceiveData(buf, len, out, &SPI2_Resources);
}
#endif


/**
  \fn          ARM_SPI_STATUS SPI_AbortTransfer (SPI_RESOURCES *ptr_spi)
  \brief       Abort current SPI transfer.
  \param[in]   ptr_spi  Pointer to SPI resources
  \return      \ref ARM_SPI_STATUS
*/
static ARM_SPI_STATUS SPI_AbortTransfer (SPI_RESOURCES *ptr_spi) {
  if (!ptr_spi->info->power) return ARM_SPI_ERROR;

  /* Disable interrupts */
  ptr_spi->usart->IEN &= ~(USART_IEN_RXDATAV | USART_IEN_TXBL);

  /* Release semaphores */
  osSemaphoreRelease(ptr_spi->info->semaphore_xfer);
  osSemaphoreWait   (ptr_spi->info->semaphore_xfer, 0);
  osSemaphoreRelease(ptr_spi->info->semaphore_busy);
  osSemaphoreWait   (ptr_spi->info->semaphore_busy, 0);
  
  return ARM_SPI_OK;
}

#if (RTE_USART0 && RTE_USART0_SYNC)
static ARM_SPI_STATUS SPI0_AbortTransfer (void) {
  return SPI_AbortTransfer(&SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
static ARM_SPI_STATUS SPI1_AbortTransfer (void) {
  return SPI_AbortTransfer(&SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
static ARM_SPI_STATUS SPI2_AbortTransfer (void) {
  return SPI_AbortTransfer(&SPI2_Resources);
}
#endif


/**
  \fn          void USART_Rx_IRQHandler (SPI_RESOURCES *ptr_spi)
  \brief       USART receive interrupt handler.
  \param[in]   ptr_spi  Pointer to SPI resources
*/
static void USART_Rx_IRQHandler (SPI_RESOURCES *ptr_spi) {
  if (ptr_spi->usart->IF & USART_IF_RXDATAV) {
    /* Data available in the receive buffer */
    if (ptr_spi->info->rx.buf) {
      *(ptr_spi->info->rx.buf) = ptr_spi->usart->RXDATA;

      if (ptr_spi->info->rx.inc) {
        ptr_spi->info->rx.buf++;
      }
    }
    else {
      ptr_spi->usart->RXDATA;
    }
    ptr_spi->info->rx.cnt--;
    
    if (ptr_spi->info->rx.cnt == 0) {
      /* Disable Rx Data Valid interrupt  */
      ptr_spi->usart->IEN &= ~USART_IEN_RXDATAV;
      NVIC_ClearPendingIRQ (ptr_spi->usart_rx_irq);

      if (ptr_spi->info->part) {
        osSemaphoreRelease (ptr_spi->info->semaphore_part);
      }
      else {
        if (ptr_spi->info->rx.buf) {
          osSemaphoreRelease (ptr_spi->info->semaphore_xfer);
        }
        else {
          if (ptr_spi->info->cb_event) {
            ptr_spi->info->cb_event (ARM_SPI_EVENT_SEND_DONE);
          }
        }
        osSemaphoreRelease (ptr_spi->info->semaphore_busy);
      }
    }
  }
}


/**
  \fn          void USART_Tx_IRQHandler (SPI_RESOURCES *ptr_spi)
  \brief       USART transmit interrupt handler.
  \param[in]   ptr_spi  Pointer to SPI resources
*/
static void USART_Tx_IRQHandler (SPI_RESOURCES *ptr_spi) {
  if (ptr_spi->usart->IF & USART_IF_TXBL) {
    /* Buffer empty or TX complete */
    if (ptr_spi->info->tx.cnt) {
      ptr_spi->usart->TXDATA = *(ptr_spi->info->tx.buf);
      if (ptr_spi->info->tx.inc) {
        ptr_spi->info->tx.buf++;
      }
      ptr_spi->info->tx.cnt--;
    }
    else {
      /* Disable TX Buffer Empty interrupt */
      ptr_spi->usart->IEN &= ~USART_IEN_TXBL;
      NVIC_ClearPendingIRQ (ptr_spi->usart_tx_irq);
      
      if (ptr_spi->info->part == false) {
        if (ptr_spi->info->rx.buf == NULL) {
          osSemaphoreRelease (ptr_spi->info->semaphore_xfer);
        }
      }
    }
  }
}


#if (RTE_USART0 && RTE_USART0_SYNC)
void USART0_RX_IRQHandler (void) {
  USART_Rx_IRQHandler(&SPI0_Resources);
}
void USART0_TX_IRQHandler (void) {
  USART_Tx_IRQHandler(&SPI0_Resources);
}
#endif

#if (RTE_USART1 && RTE_USART1_SYNC)
void USART1_RX_IRQHandler (void) {
  USART_Rx_IRQHandler(&SPI1_Resources);
}
void USART1_TX_IRQHandler (void) {
  USART_Tx_IRQHandler(&SPI1_Resources);
}
#endif

#if (RTE_USART2 && RTE_USART2_SYNC)
void USART2_RX_IRQHandler (void) {
  USART_Rx_IRQHandler(&SPI2_Resources);
}
void USART2_TX_IRQHandler (void) {
  USART_Tx_IRQHandler(&SPI2_Resources);
}
#endif


/* DMA Callbacks */

#if defined(USART_TX_DMA_ENABLE)

/**
  \fn          void USART_DMA_TxComplete (SPI_RESOURCES *ptr_spi)
  \brief       DMA transmit complete callback.
  \param[in]   ptr_spi  Pointer to SPI resources
*/
static void USART_DMA_TxComplete (SPI_RESOURCES *ptr_spi) {

  if (ptr_spi->info->part == false) {
    if (ptr_spi->info->rx.buf == NULL) {
      osSemaphoreRelease (ptr_spi->info->semaphore_xfer);
    }
  }
}
#endif


#if defined(USART_RX_DMA_ENABLE)

/**
  \fn          void USART_DMA_RxComplete (SPI_RESOURCES *ptr_spi)
  \brief       DMA receive complete callback.
  \param[in]   ptr_spi  Pointer to SPI resources
*/
static void USART_DMA_RxComplete (SPI_RESOURCES *ptr_spi) {

  if (ptr_spi->info->part) {
    /* Start sending new chunk */
    osSemaphoreRelease (ptr_spi->info->semaphore_part);
  }
  else {
    /* DMA transfered all data */
    if (ptr_spi->info->rx.buf) {
      /* Master receive completed */
      osSemaphoreRelease (ptr_spi->info->semaphore_xfer);
    }
    else {
      /* Master transmit completed */
      if (ptr_spi->info->cb_event) {
         (ptr_spi->info->cb_event)(ARM_SPI_EVENT_SEND_DONE);
      }
    }
    osSemaphoreRelease (ptr_spi->info->semaphore_busy);
  }
}
#endif


#if defined(USART_TX_DMA_ENABLE) || defined(USART_RX_DMA_ENABLE)

/**
  \fn          void USART_DMA_Complete (uint32_t channel, bool primary, void *userPtr)
  \brief       DMA transaction complete callback.
  \param[in]   channel    Channel number
  \param[in]   primary    Primary descriptor
  \param[in]   userPtr    User defined pointer
*/
void USART_DMA_Complete (uint32_t channel, bool primary, void *userPtr) {
  #if (RTE_USART0 && RTE_USART0_SYNC && RTE_USART0_TX_DMA)
  if (channel == RTE_USART0_TX_DMA_CHANNEL) {
    USART_DMA_TxComplete (&SPI0_Resources);
  }
  #endif

  #if (RTE_USART0 && RTE_USART0_SYNC && RTE_USART0_RX_DMA)
  if (channel == RTE_USART0_RX_DMA_CHANNEL) {
    USART_DMA_RxComplete (&SPI0_Resources);
  }
  #endif

  #if (RTE_USART1 && RTE_USART1_SYNC && RTE_USART1_TX_DMA)
  if (channel == RTE_USART1_TX_DMA_CHANNEL) {
    USART_DMA_TxComplete (&SPI1_Resources);
  }
  #endif

  #if (RTE_USART1 && RTE_USART1_SYNC && RTE_USART1_RX_DMA)
  if (channel == RTE_USART1_RX_DMA_CHANNEL) {
    USART_DMA_RxComplete (&SPI1_Resources);
  }
  #endif

  #if (RTE_USART2 && RTE_USART2_SYNC && RTE_USART2_TX_DMA)
  if (channel == RTE_USART2_TX_DMA_CHANNEL) {
    USART_DMA_TxComplete (&SPI2_Resources);
  }
  #endif
  
  #if (RTE_USART2 && RTE_USART2_SYNC && RTE_USART2_RX_DMA)
  if (channel == RTE_USART2_RX_DMA_CHANNEL) {
    USART_DMA_RxComplete (&SPI2_Resources);
  }
  #endif
}
#endif


/* SPI0 Driver Control Block */
#if (RTE_USART0 && RTE_USART0_SYNC)
ARM_DRIVER_SPI Driver_SPI0 = {
  SPIX_GetVersion,
  SPIX_GetCapabilities,
  SPI0_Initialize,
  SPI0_Uninitialize,
  SPI0_PowerControl,
  SPI0_Configure,
  SPI0_BusSpeed,
  SPI0_SlaveSelect,
  SPI0_TransferByte,
  SPI0_SendData,
  SPI0_ReceiveData,
  SPI0_AbortTransfer
};
#endif

/* SPI1 Driver Control Block */
#if (RTE_USART1 && RTE_USART1_SYNC)
ARM_DRIVER_SPI Driver_SPI1 = {
  SPIX_GetVersion,
  SPIX_GetCapabilities,
  SPI1_Initialize,
  SPI1_Uninitialize,
  SPI1_PowerControl,
  SPI1_Configure,
  SPI1_BusSpeed,
  SPI1_SlaveSelect,
  SPI1_TransferByte,
  SPI1_SendData,
  SPI1_ReceiveData,
  SPI1_AbortTransfer
};
#endif

/* SPI2 Driver Control Block */
#if (RTE_USART2 && RTE_USART2_SYNC)
ARM_DRIVER_SPI Driver_SPI2 = {
  SPIX_GetVersion,
  SPIX_GetCapabilities,
  SPI2_Initialize,
  SPI2_Uninitialize,
  SPI2_PowerControl,
  SPI2_Configure,
  SPI2_BusSpeed,
  SPI2_SlaveSelect,
  SPI2_TransferByte,
  SPI2_SendData,
  SPI2_ReceiveData,
  SPI2_AbortTransfer
};
#endif
