/* ----------------------------------------------------------------------
 * Copyright (C) 2013 ARM Limited. All rights reserved.
 *  
 * $Date:        13. September 2013
 * $Revision:    V1.01
 *  
 * Driver:       Driver_UART0, Driver_UART1, Driver_UART2, Driver_UART3,
 *               Driver_UART4
 * Configured:   via RTE_Device.h configuration file 
 * Project:      UART Driver for Energy Micro EFM32GGxxx
 * ---------------------------------------------------------------------- 
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 * 
 *   Configuration Setting               Value     UART Interface
 *   ---------------------               -----     -------------
 *   Connect to hardware via Driver_UART# = 0       use UART0
 *   Connect to hardware via Driver_UART# = 1       use UART1
 *   Connect to hardware via Driver_UART# = 2       use USART0
 *   Connect to hardware via Driver_UART# = 3       use USART1
 *   Connect to hardware via Driver_UART# = 4       use USART2
 * -------------------------------------------------------------------- */

/* History:
 *  Version 1.01
 *    Based on API V1.10 (namespace prefix ARM_ added)
 *  Version 1.00
 *    Initial release
 */

#include "cmsis_os.h"
#include "em_device.h"

#include "em_cmu.h"
#include "em_dma.h"
#include "em_gpio.h"
#include "em_usart.h"

#include "Driver_UART.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

#include <string.h>

#define ARM_UART_DRV_VERSION ARM_DRV_VERSION_MAJOR_MINOR(1,01)  /* driver version */

#if ((defined(RTE_Drivers_UART0)  || \
      defined(RTE_Drivers_UART1)  || \
      defined(RTE_Drivers_UART2) || \
      defined(RTE_Drivers_UART3) || \
      defined(RTE_Drivers_UART4))   \
     && !RTE_UART0                   \
     && !RTE_UART1                   \
     && !RTE_USART0                  \
     && !RTE_USART1                  \
     && !RTE_USART2)
#error "UART not configured in RTE_Device.h!"
#endif

#if (RTE_UART0 && RTE_UART0_RX_DMA)
  #warning "UART0 driver does not support receive DMA! Disable receive DMA in RTE_Device.h!"
#endif

#if (RTE_UART1 && RTE_UART1_RX_DMA)
  #warning "UART1 driver does not support receive DMA! Disable receive DMA in RTE_Device.h!"
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0) && RTE_USART0_RX_DMA)
  #warning "UART2 driver does not support receive DMA! Disable receive DMA in RTE_Device.h!"
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0) && RTE_USART1_RX_DMA)
  #warning "UART3 driver does not support receive DMA! Disable receive DMA in RTE_Device.h!"
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0)  && RTE_USART2_RX_DMA)
  #warning "UART4 driver does not support receive DMA! Disable receive DMA in RTE_Device.h!"
#endif

#if ((RTE_UART0  && RTE_UART0_TX_DMA)                            || \
     (RTE_UART1  && RTE_UART1_TX_DMA)                            || \
     (RTE_USART0 && (RTE_USART0_SYNC == 0) && RTE_USART0_TX_DMA) || \
     (RTE_USART1 && (RTE_USART1_SYNC == 0) && RTE_USART1_TX_DMA) || \
     (RTE_USART2 && (RTE_USART2_SYNC == 0) && RTE_USART2_TX_DMA))
  #define UART_TX_DMA_ENABLE
#endif

/* UART configuration flags */
#define UART_CFG_TX_DMA_ENABLE          (1U << 0)

/* DMA channel request signal */
#define DMA_CH_CTRL_SIGSEL_USARTTXEMPTY (2)

/* USART Transfer Timeout in microseconds */
#define USART_TRANSFER_TIMEOUT          (500000)

/* Driver Version */
static const ARM_DRV_VERSION DriverVersion = {
  ARM_UART_API_VERSION,
  ARM_UART_DRV_VERSION
};

/* UART Data Information (Run-Time) */
typedef struct _UART_DATA_INFO {
  volatile uint16_t      idx_in;
  volatile uint16_t      idx_out;
  volatile bool          empty;
           uint16_t      threshold;
} UART_DATA_INFO;

/* UART Information (Run-Time) */
typedef struct _UART_INFO {
  bool                   power;
  ARM_UART_SignalEvent_t cb_event;
  uint32_t               event_mask;
  osSemaphoreId          semaphore;
  osMutexId              mutex_write;
  osMutexId              mutex_read;
  UART_DATA_INFO         rx;
  UART_DATA_INFO         tx;
  uint16_t               dma_xfer_len;
  bool                   com_break;
} UART_INFO;

/* GPIO pin description */
typedef struct _GPIO_PIN_ID {
  GPIO_Port_TypeDef      port;
  uint8_t                num;
} GPIO_PIN_ID;

/* UART Resources definition */
typedef struct _UART_RESOURCES {
  const ARM_UART_CAPABILITIES *capabilities;    // Capabilities
        GPIO_PIN_ID            pin_tx;          // TX  Pin identifier
        GPIO_PIN_ID            pin_rx;          // RX  Pin identifier
        USART_TypeDef         *uart;            // Pointer to UART peripheral
  
  enum  IRQn                   uart_rx_irq;     // UART Rx IRQ Number
  enum  IRQn                   uart_tx_irq;     // UART Tx IRQ Number

        CMU_Clock_TypeDef      clock_mask;      // Clock register mask

        uint8_t               *tx_buffer;       // Transmit buffer
        uint32_t               tx_buffer_size;  // Transmit buffer size
        uint8_t               *rx_buffer;       // Receive buffer
        uint32_t               rx_buffer_size;  // Receive buffer size

        uint8_t                dma_tx_ch;       // UART Transmit DMA channel
        uint32_t               dma_sourcesel;   // DMA channel input source
        
        uint8_t                pin_loc;         // Pin location
        uint8_t                cfg;             // Driver configuration

  const osSemaphoreDef_t      *semaphore;       // Tx Semaphore definition
  const osMutexDef_t          *mutex_write;     // Write Mutex definition
  const osMutexDef_t          *mutex_read;      // Write Mutex definition
        UART_INFO             *info;            // Run-Time information
} const UART_RESOURCES;

#if defined(UART_TX_DMA_ENABLE)
/* DMA Descriptor array */
extern DMA_DESCRIPTOR_TypeDef DMA_DescArray[];

/* DMA Callback function */
void UART_DMA_Complete (uint32_t channel, bool primary, void *ptr_uart);

/* DMA Callback */
DMA_CB_TypeDef UART_DMA_Cb = {
  UART_DMA_Complete,
  NULL,
  0
};
#endif

/* UART0 */
#if (RTE_UART0)

#ifndef __UART0_TX_BUFFER_SIZE          /* Buffer size must be 2^n            */
#define __UART0_TX_BUFFER_SIZE          128
#endif

#ifndef __UART0_RX_BUFFER_SIZE          /* Buffer size must be 2^n            */
#define __UART0_RX_BUFFER_SIZE          128
#endif

/* UART0 Capabilities */
static const ARM_UART_CAPABILITIES UART0_Capabilities = {
  __UART0_TX_BUFFER_SIZE,               // Transmit Buffer Size
  __UART0_RX_BUFFER_SIZE,               // Receive Buffer Size
  0,                                    // Receive character timeout event
  0,                                    // CTS Line available
  0,                                    // Signal CTS change event
  0,                                    // DSR Line available
  0,                                    // Signal DSR change event
  0,                                    // DCD Line available
  0,                                    // Signal DCD change event
  0,                                    // RI Line available
  0,                                    // Signal RI change event
  0
};

/* UART0 Semaphore */
osSemaphoreDef(UART0_Semaphore);
osSemaphoreId  UART0_Semaphore;

/* UART0 Mutex */
osMutexDef(UART0_Mutex_write);
osMutexDef(UART0_Mutex_read);

/* UART0 Data Info */
struct {
  uint8_t  tx[__UART0_TX_BUFFER_SIZE];
  uint8_t  rx[__UART0_RX_BUFFER_SIZE];
} UART0_Buffer;

/* UART0 Information (Run-Time) */
static UART_INFO UART0_Info;

/* UART0 Resources */
static UART_RESOURCES UART0_Resources = {
  &UART0_Capabilities,
  { RTE_UART0_TX_PORT,  RTE_UART0_TX_BIT  },
  { RTE_UART0_RX_PORT,  RTE_UART0_RX_BIT  },
  UART0,                                /* Pointer to UART peripheral         */
  UART0_RX_IRQn,                        /* UART Rx IRQ Number                 */
  UART0_TX_IRQn,                        /* UART Tx IRQ Number                 */
  
  cmuClock_UART0,                       /* Clock register bit mask            */
  
  UART0_Buffer.tx,                      /* Tx buffer address                  */
  __UART0_TX_BUFFER_SIZE,               /* Tx buffer size                     */
  UART0_Buffer.rx,                      /* Rx buffer address                  */
  __UART0_RX_BUFFER_SIZE,               /* Rx buffer size                     */

  RTE_UART0_TX_DMA_CHANNEL,             /* UART transmit DMA channel          */
  DMA_CH_CTRL_SOURCESEL_UART0,          /* DMA channel source peripheral      */

  RTE_UART0_PIN_LOC,                    /* UART pin location                  */

  #if (RTE_UART0_TX_DMA)
  UART_CFG_TX_DMA_ENABLE |              /* UART transmit DMA enable           */
  #endif
  0,

  osSemaphore(UART0_Semaphore),         /* UART synchonisation semaphore      */
  osMutex(UART0_Mutex_write),           /* UART write mutex                   */
  osMutex(UART0_Mutex_read),            /* UART read mutex                    */
  &UART0_Info                           /* UART runtime information structure */
};
#endif /* RTE_UART0 */


/* UART1 */
#if (RTE_UART1)

#ifndef __UART1_TX_BUFFER_SIZE          /* Buffer size must be 2^n            */
#define __UART1_TX_BUFFER_SIZE          128
#endif

#ifndef __UART1_RX_BUFFER_SIZE          /* Buffer size must be 2^n            */
#define __UART1_RX_BUFFER_SIZE          128
#endif

/* UART1 Capabilities */
static const ARM_UART_CAPABILITIES UART1_Capabilities = {
  __UART1_TX_BUFFER_SIZE,               // Transmit Buffer Size
  __UART1_RX_BUFFER_SIZE,               // Receive Buffer Size
  0,                                    // Receive character timeout event
  0,                                    // CTS Line available
  0,                                    // Signal CTS change event
  0,                                    // DSR Line available
  0,                                    // Signal DSR change event
  0,                                    // DCD Line available
  0,                                    // Signal DCD change event
  0,                                    // RI Line available
  0,                                    // Signal RI change event
  0
};

/* UART1 Semaphore */
osSemaphoreDef(UART1_Semaphore);
osSemaphoreId  UART1_Semaphore;

/* UART1 Mutex */
osMutexDef(UART1_Mutex_write);
osMutexDef(UART1_Mutex_read);

/* UART1 Data Info */
struct {
  uint8_t  tx[__UART1_TX_BUFFER_SIZE];
  uint8_t  rx[__UART1_RX_BUFFER_SIZE];
} UART1_Buffer;

/* UART1 Information (Run-Time) */
static UART_INFO UART1_Info;

/* UART1 Resources */
static UART_RESOURCES UART1_Resources = {
  &UART1_Capabilities,
  { RTE_UART1_TX_PORT,  RTE_UART1_TX_BIT  },
  { RTE_UART1_RX_PORT,  RTE_UART1_RX_BIT  },
  UART1,                                /* Pointer to UART peripheral         */
  UART1_RX_IRQn,                        /* UART Rx IRQ Number                 */
  UART1_TX_IRQn,                        /* UART Tx IRQ Number                 */
  
  cmuClock_UART1,                       /* Clock register bit mask            */
  
  UART1_Buffer.tx,                      /* Tx buffer address                  */
  __UART1_TX_BUFFER_SIZE,               /* Tx buffer size                     */
  UART1_Buffer.rx,                      /* Rx buffer address                  */
  __UART1_RX_BUFFER_SIZE,               /* Rx buffer size                     */

  RTE_UART1_TX_DMA_CHANNEL,             /* UART transmit DMA channel          */
  DMA_CH_CTRL_SOURCESEL_UART1,          /* DMA channel source peripheral      */

  RTE_UART1_PIN_LOC,                    /* UART pin location                  */

  #if (RTE_UART1_TX_DMA)
  UART_CFG_TX_DMA_ENABLE |              /* UART transmit DMA enable           */
  #endif
  0,

  osSemaphore(UART1_Semaphore),         /* UART synchonisation semaphore      */
  osMutex(UART1_Mutex_write),           /* UART write mutex                   */
  osMutex(UART1_Mutex_read),            /* UART read mutex                    */
  &UART1_Info                           /* UART runtime information structure */
};
#endif /* RTE_UART1 */

/* USART0 */
#if (RTE_USART0 && (RTE_USART0_SYNC == 0))

#ifndef __USART0_TX_BUFFER_SIZE         /* Buffer size must be 2^n            */
#define __USART0_TX_BUFFER_SIZE         128
#endif

#ifndef __USART0_RX_BUFFER_SIZE         /* Buffer size must be 2^n            */
#define __USART0_RX_BUFFER_SIZE         128
#endif

/* USART0 Capabilities */
static const ARM_UART_CAPABILITIES USART0_Capabilities = {
  __USART0_TX_BUFFER_SIZE,              // Transmit Buffer Size
  __USART0_RX_BUFFER_SIZE,              // Receive Buffer Size
  0,                                    // Receive character timeout event
  0,                                    // CTS Line available
  0,                                    // Signal CTS change event
  0,                                    // DSR Line available
  0,                                    // Signal DSR change event
  0,                                    // DCD Line available
  0,                                    // Signal DCD change event
  0,                                    // RI Line available
  0,                                    // Signal RI change event
  0
};

/* USART0 Semaphore */
osSemaphoreDef(USART0_Semaphore);
osSemaphoreId  USART0_Semaphore;

/* USART0 Mutex */
osMutexDef(USART0_Mutex_write);
osMutexDef(USART0_Mutex_read);

/* USART0 Data Info */
struct {
  uint8_t  tx[__USART0_TX_BUFFER_SIZE];
  uint8_t  rx[__USART0_RX_BUFFER_SIZE];
} USART0_Buffer;

/* USART0 Information (Run-Time) */
static UART_INFO USART0_Info;

/* USART0 Resources */
static UART_RESOURCES USART0_Resources = {
  &USART0_Capabilities,
  { RTE_USART0_TX_PORT,  RTE_USART0_TX_BIT  },
  { RTE_USART0_RX_PORT,  RTE_USART0_RX_BIT  },
  USART0,                               /* Pointer to UART peripheral         */
  USART0_RX_IRQn,                       /* UART Rx IRQ Number                 */
  USART0_TX_IRQn,                       /* UART Tx IRQ Number                 */
  
  cmuClock_USART0,                      /* Clock register bit mask            */
  
  USART0_Buffer.tx,                     /* Tx buffer address                  */
  __USART0_TX_BUFFER_SIZE,              /* Tx buffer size                     */
  USART0_Buffer.rx,                     /* Rx buffer address                  */
  __USART0_RX_BUFFER_SIZE,              /* Rx buffer size                     */

  RTE_USART0_TX_DMA_CHANNEL,            /* UART transmit DMA channel          */
  DMA_CH_CTRL_SOURCESEL_USART0,         /* DMA channel source peripheral      */

  RTE_USART0_PIN_LOC,                   /* UART pin location                  */

  #if (RTE_USART0_TX_DMA)
  UART_CFG_TX_DMA_ENABLE |              /* UART transmit DMA enable           */
  #endif
  0,

  osSemaphore(USART0_Semaphore),        /* UART synchonisation semaphore      */
  osMutex(USART0_Mutex_write),          /* UART write mutex                   */
  osMutex(USART0_Mutex_read),           /* UART read mutex                    */
  &USART0_Info                          /* UART runtime information structure */
};
#endif /* RTE_USART0 && (RTE_USART0_SYNC == 0) */

/* USART1 */
#if (RTE_USART1 && (RTE_USART1_SYNC == 0))

#ifndef __USART1_TX_BUFFER_SIZE         /* Buffer size must be 2^n            */
#define __USART1_TX_BUFFER_SIZE         128
#endif

#ifndef __USART1_RX_BUFFER_SIZE         /* Buffer size must be 2^n            */
#define __USART1_RX_BUFFER_SIZE         128
#endif

/* USART1 Capabilities */
static const ARM_UART_CAPABILITIES USART1_Capabilities = {
  __USART1_TX_BUFFER_SIZE,              // Transmit Buffer Size
  __USART1_RX_BUFFER_SIZE,              // Receive Buffer Size
  0,                                    // Receive character timeout event
  0,                                    // CTS Line available
  0,                                    // Signal CTS change event
  0,                                    // DSR Line available
  0,                                    // Signal DSR change event
  0,                                    // DCD Line available
  0,                                    // Signal DCD change event
  0,                                    // RI Line available
  0,                                    // Signal RI change event
  0
};

/* USART1 Semaphore */
osSemaphoreDef(USART1_Semaphore);
osSemaphoreId  USART1_Semaphore;

/* USART1 Mutex */
osMutexDef(USART1_Mutex_write);
osMutexDef(USART1_Mutex_read);

/* USART1 Data Info */
struct {
  uint8_t  tx[__USART1_TX_BUFFER_SIZE];
  uint8_t  rx[__USART1_RX_BUFFER_SIZE];
} USART1_Buffer;

/* USART1 Information (Run-Time) */
static UART_INFO USART1_Info;

/* USART1 Resources */
static UART_RESOURCES USART1_Resources = {
  &USART1_Capabilities,
  { RTE_USART1_TX_PORT,  RTE_USART1_TX_BIT  },
  { RTE_USART1_RX_PORT,  RTE_USART1_RX_BIT  },
  USART1,                               /* Pointer to UART peripheral         */
  USART1_RX_IRQn,                       /* UART Rx IRQ Number                 */
  USART1_TX_IRQn,                       /* UART Tx IRQ Number                 */
  
  cmuClock_USART1,                      /* Clock register bit mask            */
  
  USART1_Buffer.tx,                     /* Tx buffer address                  */
  __USART1_TX_BUFFER_SIZE,              /* Tx buffer size                     */
  USART1_Buffer.rx,                     /* Rx buffer address                  */
  __USART1_RX_BUFFER_SIZE,              /* Rx buffer size                     */

  RTE_USART1_TX_DMA_CHANNEL,            /* UART transmit DMA channel          */
  DMA_CH_CTRL_SOURCESEL_USART1,         /* DMA channel source peripheral      */

  RTE_USART1_PIN_LOC,                   /* UART pin location                  */

  #if (RTE_USART1_TX_DMA)
  UART_CFG_TX_DMA_ENABLE |              /* UART transmit DMA enable           */
  #endif
  0,

  osSemaphore(USART1_Semaphore),        /* UART synchonisation semaphore      */
  osMutex(USART1_Mutex_write),          /* UART write mutex                   */
  osMutex(USART1_Mutex_read),           /* UART read mutex                    */
  &USART1_Info                          /* UART runtime information structure */
};
#endif /* RTE_USART1 && (RTE_USART1_SYNC == 0) */

/* USART2 */
#if (RTE_USART2 && (RTE_USART2_SYNC == 0))

#ifndef __USART2_TX_BUFFER_SIZE         /* Buffer size must be 2^n            */
#define __USART2_TX_BUFFER_SIZE         128
#endif

#ifndef __USART2_RX_BUFFER_SIZE         /* Buffer size must be 2^n            */
#define __USART2_RX_BUFFER_SIZE         128
#endif

/* USART2 Capabilities */
static const ARM_UART_CAPABILITIES USART2_Capabilities = {
  __USART2_TX_BUFFER_SIZE,              // Transmit Buffer Size
  __USART2_RX_BUFFER_SIZE,              // Receive Buffer Size
  0,                                    // Receive character timeout event
  0,                                    // CTS Line available
  0,                                    // Signal CTS change event
  0,                                    // DSR Line available
  0,                                    // Signal DSR change event
  0,                                    // DCD Line available
  0,                                    // Signal DCD change event
  0,                                    // RI Line available
  0,                                    // Signal RI change event
  0
};

/* USART2 Semaphore */
osSemaphoreDef(USART2_Semaphore);
osSemaphoreId  USART2_Semaphore;

/* USART2 Mutex */
osMutexDef(USART2_Mutex_write);
osMutexDef(USART2_Mutex_read);

/* USART2 Data Info */
struct {
  uint8_t  tx[__USART2_TX_BUFFER_SIZE];
  uint8_t  rx[__USART2_RX_BUFFER_SIZE];
} USART2_Buffer;

/* USART2 Information (Run-Time) */
static UART_INFO USART2_Info;

/* USART2 Resources */
static UART_RESOURCES USART2_Resources = {
  &USART2_Capabilities,
  { RTE_USART2_TX_PORT,  RTE_USART2_TX_BIT  },
  { RTE_USART2_RX_PORT,  RTE_USART2_RX_BIT  },
  USART2,                               /* Pointer to UART peripheral         */
  USART2_RX_IRQn,                       /* UART Rx IRQ Number                 */
  USART2_TX_IRQn,                       /* UART Tx IRQ Number                 */
  
  cmuClock_USART2,                      /* Clock register bit mask            */
  
  USART2_Buffer.tx,                     /* Tx buffer address                  */
  __USART2_TX_BUFFER_SIZE,              /* Tx buffer size                     */
  USART2_Buffer.rx,                     /* Rx buffer address                  */
  __USART2_RX_BUFFER_SIZE,              /* Rx buffer size                     */

  RTE_USART2_TX_DMA_CHANNEL,            /* UART transmit DMA channel          */
  DMA_CH_CTRL_SOURCESEL_USART1,         /* DMA channel source peripheral      */

  RTE_USART2_PIN_LOC,                   /* UART pin location                  */

  #if (RTE_USART2_TX_DMA)
  UART_CFG_TX_DMA_ENABLE |              /* UART transmit DMA enable           */
  #endif
  0,

  osSemaphore(USART2_Semaphore),        /* UART synchonisation semaphore      */
  osMutex(USART2_Mutex_write),          /* UART write mutex                   */
  osMutex(USART2_Mutex_read),           /* UART read mutex                    */
  &USART2_Info                          /* UART runtime information structure */
};
#endif /* RTE_USART2 && (RTE_USART2_SYNC == 0) */


/**
  \fn          ARM_DRV_VERSION UARTx_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRV_VERSION UARTX_GetVersion (void) {
  return DriverVersion;
}


/**
  \fn          ARM_UART_CAPABILITIES UART_GetCapabilities (UART_RESOURCES *ptr_uart)
  \brief       Get driver capabilities.
  \param[in]   ptr_uart    Pointer to UART resources
  \return      \ref ARM_UART_CAPABILITIES
*/
static ARM_UART_CAPABILITIES UART_GetCapabilities (UART_RESOURCES *ptr_uart) {
  return *ptr_uart->capabilities;
}

#if (RTE_UART0)
  static ARM_UART_CAPABILITIES UART0_GetCapabilities (void) {
    return UART_GetCapabilities (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_CAPABILITIES UART1_GetCapabilities (void) {
    return UART_GetCapabilities (&UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_CAPABILITIES USART0_GetCapabilities (void) {
    return UART_GetCapabilities (&USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_CAPABILITIES USART1_GetCapabilities (void) {
    return UART_GetCapabilities (&USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_CAPABILITIES USART2_GetCapabilities (void) {
    return UART_GetCapabilities (&USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_Initialize (ARM_UART_SignalEvent_t cb_event,
                                                uint32_t               event_mask
                                                UART_RESOURCES        *ptr_uart)
  \brief       Initialize UART Interface.
  \param[in]   cb_event    Pointer to \ref ARM_UART_SignalEvent
  \param[in]   event_mask
  \param[in]   ptr_uart    Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_Initialize (ARM_UART_SignalEvent_t cb_event,
                                        uint32_t               event_mask,
                                        UART_RESOURCES        *ptr_uart) {

  #if defined(UART_TX_DMA_ENABLE)
  DMA_Init_TypeDef DmaInit;
  #endif

  /* Enable UARTx peripheral clock */
  CMU_ClockEnable (ptr_uart->clock_mask, true);

  /* Enable GPIO peripheral clock */
  CMU_ClockEnable (cmuClock_GPIO, true);

  /* Configure UART TX pin */
  GPIO_PinModeSet (ptr_uart->pin_tx.port, ptr_uart->pin_tx.num, gpioModePushPull, 0);
  /* Configure UART RX pin */
  GPIO_PinModeSet (ptr_uart->pin_rx.port, ptr_uart->pin_rx.num, gpioModePushPull, 0);

  /* Define USARTx pin location and enable Tx, Rx, CLK and CS pins */
  ptr_uart->uart->ROUTE = (ptr_uart->uart->ROUTE & ~_USART_ROUTE_LOCATION_MASK)  |
                          (ptr_uart->pin_loc     << _USART_ROUTE_LOCATION_SHIFT) |
                           USART_ROUTE_RXPEN | USART_ROUTE_TXPEN;

  #if defined(UART_TX_DMA_ENABLE)
  if (ptr_uart->cfg & UART_CFG_TX_DMA_ENABLE) {
    DmaInit.hprot        = false;
    DmaInit.controlBlock = DMA_DescArray;
    DMA_Init (&DmaInit);
  }
  #endif

  /* Initialize Run-Time Resources */
  ptr_uart->info->power           = false;
  ptr_uart->info->cb_event        = cb_event;
  ptr_uart->info->event_mask      = event_mask;

  ptr_uart->info->tx.idx_in       = 0;
  ptr_uart->info->tx.idx_out      = 0;
  ptr_uart->info->tx.empty        = true;

  ptr_uart->info->rx.idx_in       = 0;
  ptr_uart->info->rx.idx_out      = 0;
  ptr_uart->info->rx.empty        = true;

  ptr_uart->info->com_break       = false;

  if (ptr_uart->info->semaphore   == NULL)
    ptr_uart->info->semaphore     = osSemaphoreCreate(ptr_uart->semaphore, 1);
  if (ptr_uart->info->mutex_write == NULL)
    ptr_uart->info->mutex_write   = osMutexCreate(ptr_uart->mutex_write);
  if (ptr_uart->info->mutex_read  == NULL)
    ptr_uart->info->mutex_read    = osMutexCreate(ptr_uart->mutex_read);

  if (ptr_uart->info->semaphore   == NULL) return ARM_UART_ERROR;
  if (ptr_uart->info->mutex_write == NULL) return ARM_UART_ERROR;
  if (ptr_uart->info->mutex_read  == NULL) return ARM_UART_ERROR;

  /* Enable USART interrupts */
  NVIC_ClearPendingIRQ (ptr_uart->uart_tx_irq);
  NVIC_ClearPendingIRQ (ptr_uart->uart_rx_irq);
  NVIC_EnableIRQ (ptr_uart->uart_tx_irq);
  NVIC_EnableIRQ (ptr_uart->uart_rx_irq);

  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_Initialize (ARM_UART_SignalEvent_t cb_event,
                                           uint32_t               event_mask) {
    return UART_Initialize (cb_event, event_mask, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_Initialize (ARM_UART_SignalEvent_t cb_event,
                                           uint32_t               event_mask) {
    return UART_Initialize (cb_event, event_mask, &UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_STATUS USART0_Initialize (ARM_UART_SignalEvent_t cb_event,
                                            uint32_t               event_mask) {
    return UART_Initialize (cb_event, event_mask, &USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_STATUS USART1_Initialize (ARM_UART_SignalEvent_t cb_event,
                                            uint32_t               event_mask) {
    return UART_Initialize (cb_event, event_mask, &USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_STATUS USART2_Initialize (ARM_UART_SignalEvent_t cb_event,
                                            uint32_t               event_mask) {
    return UART_Initialize (cb_event, event_mask, &USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_Uninitialize (UART_RESOURCES *ptr_uart)
  \brief       De-initialize UART Interface.
  \param[in]   ptr_uart  Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_Uninitialize (UART_RESOURCES *ptr_uart) {

  /* Enable clock */
  CMU_ClockEnable (ptr_uart->clock_mask, true);

  /* Disable peripheral IRQs in NVIC */
  NVIC_DisableIRQ (ptr_uart->uart_tx_irq);
  NVIC_DisableIRQ (ptr_uart->uart_rx_irq);

  /* Disable UART pins */
  ptr_uart->uart->ROUTE = 0;

  /* Disable UART Clock */
  CMU_ClockEnable (ptr_uart->clock_mask, false);

  /* Unconfigure UART Tx and Rx pins */
  GPIO_PinModeSet (ptr_uart->pin_tx.port, ptr_uart->pin_tx.num, gpioModeDisabled, 0);
  GPIO_PinModeSet (ptr_uart->pin_rx.port, ptr_uart->pin_rx.num, gpioModeDisabled, 0);

  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_Uninitialize (void) {
    return UART_Uninitialize (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_Uninitialize (void) {
    return UART_Uninitialize (&UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_STATUS USART0_Uninitialize (void) {
    return UART_Uninitialize (&USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_STATUS USART1_Uninitialize (void) {
    return UART_Uninitialize (&USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_STATUS USART2_Uninitialize (void) {
    return UART_Uninitialize (&USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_PowerControl (ARM_POWER_STATE  state,
                                                  UART_RESOURCES  *ptr_uart)
  \brief       Controls UART Interface Power.
  \param[in]   state    Power state
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_PowerControl (ARM_POWER_STATE state, UART_RESOURCES *ptr_uart) {
  #if defined(UART_TX_DMA_ENABLE)
  DMA_CfgChannel_TypeDef DmaCfg;
  #endif

  switch (state) {
    case ARM_POWER_OFF:
      ptr_uart->info->power = false;

      /* Disable UART peripheral clock */
      CMU_ClockEnable (ptr_uart->clock_mask, false);
      break;

    case ARM_POWER_LOW:
      break;

    case ARM_POWER_FULL:
      /* Enable UART peripheral clock */
      CMU_ClockEnable (ptr_uart->clock_mask, true);

      #if defined(UART_TX_DMA_ENABLE)
      /* Initialize DMA channels */
      if (ptr_uart->cfg & UART_CFG_TX_DMA_ENABLE) {
        /* Setup DMA channel configuration structure */
        DmaCfg.cb        = &UART_DMA_Cb;
        DmaCfg.enableInt = true;
        DmaCfg.highPri   = false;
        DmaCfg.select    = ptr_uart->dma_sourcesel | DMA_CH_CTRL_SIGSEL_USARTTXEMPTY;

        /* Configure Tx channel */
        DMA_CfgChannel (ptr_uart->dma_tx_ch, &DmaCfg);
      }
      #endif

      /* Clear peripheral buffers and shift registers */
      ptr_uart->uart->CMD = UART_CMD_CLEARRX | UART_CMD_CLEARTX;

      /* Configure operation mode, oversampling, etc. */
      ptr_uart->uart->CTRL = 0;

      /* Enable rx data valid interrupt */
      ptr_uart->uart->IEN |= USART_IEN_RXDATAV;

      /* Enable Rx and Tx */
      ptr_uart->uart->CMD = USART_CMD_RXEN | USART_CMD_TXEN;

      /* Ready for operation */
      ptr_uart->info->power = true;
      break;
  }
  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_PowerControl (ARM_POWER_STATE state) {
    return UART_PowerControl (state, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_PowerControl (ARM_POWER_STATE state) {
    return UART_PowerControl (state, &UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_STATUS USART0_PowerControl (ARM_POWER_STATE state) {
    return UART_PowerControl (state, &USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_STATUS USART1_PowerControl (ARM_POWER_STATE state) {
    return UART_PowerControl (state, &USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_STATUS USART2_PowerControl (ARM_POWER_STATE state) {
    return UART_PowerControl (state, &USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_Configure (uint32_t               baudrate,
                                               uint8_t                data_bits,
                                               ARM_UART_PARITY        parity,
                                               ARM_UART_STOP_BITS     stop_bits,
                                               ARM_UART_FLOW_CONTROL  flow_control,
                                               UART_RESOURCES        *ptr_uart)
  \brief       Configure UART Interface.
  \param[in]   baudrate      Requested baudrate in bits/s
  \param[in]   data_bits     Number of data bits
  \param[in]   parity        \ref ARM_UART_PARITY
  \param[in]   stop_bits     \ref ARM_UART_STOP_BITS
  \param[in]   flow_control  \ref ARM_UART_FLOW_CONTROL
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_Configure (uint32_t               baudrate,
                                       uint8_t                data_bits,
                                       ARM_UART_PARITY        parity,
                                       ARM_UART_STOP_BITS     stop_bits,
                                       ARM_UART_FLOW_CONTROL  flow_control,
                                       UART_RESOURCES        *ptr_uart) {
  uint32_t cfg;
  uint32_t br;
  ARM_UART_STATUS rtv = ARM_UART_OK;
  
  if (ptr_uart->info->power == false) { return ARM_UART_ERROR; }

  /* Disable transmitter and receiver */
  ptr_uart->uart->CMD = USART_CMD_TXDIS | USART_CMD_RXDIS;
  
  /* Configure baudrate */
  USART_BaudrateAsyncSet (ptr_uart->uart, 0, baudrate, usartOVS16);
  
  br = USART_BaudrateGet (ptr_uart->uart);
  if (br != baudrate) {
    rtv = ARM_UART_ERROR;
  }

  /* Configure data bits */
  cfg = (data_bits - 3) & _USART_FRAME_DATABITS_MASK;

  /* Configure parity */
  switch (parity) {
    case ARM_UART_PARITY_NONE:                                  break;
    case ARM_UART_PARITY_ODD:  cfg |= USART_FRAME_PARITY_ODD;   break;
    case ARM_UART_PARITY_EVEN: cfg |= USART_FRAME_PARITY_EVEN;  break;

    default: rtv = ARM_UART_ERROR;
  }
  
  /* Configure stop bits */
  switch (stop_bits) {
    case ARM_UART_STOP_BITS_1:   cfg |= USART_FRAME_STOPBITS_ONE;         break;
    case ARM_UART_STOP_BITS_1_5: cfg |= USART_FRAME_STOPBITS_ONEANDAHALF; break;
    case ARM_UART_STOP_BITS_2:   cfg |= USART_FRAME_STOPBITS_TWO;         break;
    
    default: rtv = ARM_UART_ERROR;
  }
  
  /* Configure flow control */
  if (flow_control != ARM_UART_FLOW_CONTROL_NONE) {
    rtv = ARM_UART_ERROR;
  }
  ptr_uart->uart->FRAME = cfg;

  /* Enable transmitter and receiver */
  ptr_uart->uart->CMD = USART_CMD_TXEN | USART_CMD_RXEN;

  return rtv;
}

#if (RTE_UART0)
static ARM_UART_STATUS UART0_Configure (uint32_t               baudrate,
                                        uint8_t                data_bits,
                                        ARM_UART_PARITY        parity,
                                        ARM_UART_STOP_BITS     stop_bits,
                                        ARM_UART_FLOW_CONTROL  flow_control) {
  return UART_Configure (baudrate, data_bits, parity,
                         stop_bits, flow_control, &UART0_Resources);
}
#endif

#if (RTE_UART1)
static ARM_UART_STATUS UART1_Configure (uint32_t               baudrate,
                                        uint8_t                data_bits,
                                        ARM_UART_PARITY        parity,
                                        ARM_UART_STOP_BITS     stop_bits,
                                        ARM_UART_FLOW_CONTROL  flow_control) {
  return UART_Configure (baudrate, data_bits, parity,
                         stop_bits, flow_control, &UART1_Resources);
}
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
static ARM_UART_STATUS USART0_Configure (uint32_t               baudrate,
                                         uint8_t                data_bits,
                                         ARM_UART_PARITY        parity,
                                         ARM_UART_STOP_BITS     stop_bits,
                                         ARM_UART_FLOW_CONTROL  flow_control) {
  return UART_Configure (baudrate, data_bits, parity,
                         stop_bits, flow_control, &USART0_Resources);
}
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
static ARM_UART_STATUS USART1_Configure (uint32_t               baudrate,
                                         uint8_t                data_bits,
                                         ARM_UART_PARITY        parity,
                                         ARM_UART_STOP_BITS     stop_bits,
                                         ARM_UART_FLOW_CONTROL  flow_control) {
  return UART_Configure (baudrate, data_bits, parity,
                         stop_bits, flow_control, &USART1_Resources);
}
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
static ARM_UART_STATUS USART2_Configure (uint32_t               baudrate,
                                         uint8_t                data_bits,
                                         ARM_UART_PARITY        parity,
                                         ARM_UART_STOP_BITS     stop_bits,
                                         ARM_UART_FLOW_CONTROL  flow_control) {
  return UART_Configure (baudrate, data_bits, parity,
                         stop_bits, flow_control, &USART2_Resources);
}
#endif


/**
  \fn          int32_t UART_WriteData (const uint8_t   *data,
                                       uint32_t         size,
                                       UART_RESOURCES  *ptr_uart)
  \brief       Write data to UART transmitter.
  \param[in]   data  Pointer to buffer with data to write to UART transmitter
  \param[in]   size  Data buffer size in bytes
  \param[in]   ptr_uart Pointer to UART resources
  \return      number of data bytes written, error code if negative
*/
static int32_t UART_WriteData (const uint8_t   *data, 
                                     uint32_t   size,
                               UART_RESOURCES  *ptr_uart) {
  uint32_t cnt, offs;
  int32_t  wr_cnt = 0;
  
  if (ptr_uart->info->power == false) { return wr_cnt; }

  osMutexWait(ptr_uart->info->mutex_write, osWaitForever);

  /* Determine number of data bytes in buffer */
  cnt = (ptr_uart->info->tx.idx_in - ptr_uart->info->tx.idx_out) & (ptr_uart->tx_buffer_size - 1);

  /* Check if free space in the tx buffer */
  if (size && !((cnt == 0) && !ptr_uart->info->tx.empty)) {

    /* Buffer not full - get available space in it */
    cnt = ptr_uart->tx_buffer_size - cnt;

    if (size > cnt) {
      /* Max cnt of data can be written */
      size = cnt;
    }

    offs = ptr_uart->info->tx.idx_in;

    if (size > (ptr_uart->tx_buffer_size - offs)) {
      /* Write data from in to the end of tx buffer */
      wr_cnt = ptr_uart->tx_buffer_size - offs;
      memcpy(&ptr_uart->tx_buffer[offs], data, wr_cnt);
      
      /* Adjust number of data to write and address offset */
      size -= wr_cnt;
      offs  = 0;
    }

    memcpy(&ptr_uart->tx_buffer[offs], &data[wr_cnt], size);
    wr_cnt += size;

    ptr_uart->info->tx.idx_in = (ptr_uart->info->tx.idx_in + wr_cnt) & (ptr_uart->tx_buffer_size - 1);
    ptr_uart->info->tx.empty  = false;

    if (ptr_uart->cfg & UART_CFG_TX_DMA_ENABLE) {
      if (DMA_ChannelEnabled (ptr_uart->dma_tx_ch) == false) {
        ptr_uart->info->dma_xfer_len = 0;
        /* Trigger DMA channel complete interrupt request */
        DMA->IFS = (1 << ptr_uart->dma_tx_ch);
      }
    }
    else {
      /* Enable tx buffer empty interrupt */
      ptr_uart->uart->IEN |= USART_IEN_TXBL;
    }
  }
  osMutexRelease(ptr_uart->info->mutex_write);

  return wr_cnt;
}

#if (RTE_UART0)
  static int32_t UART0_WriteData (const uint8_t *data, uint32_t size) {
    return UART_WriteData (data, size, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static int32_t UART1_WriteData (const uint8_t *data, uint32_t size) {
    return UART_WriteData (data, size, &UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static int32_t USART0_WriteData (const uint8_t *data, uint32_t size) {
    return UART_WriteData (data, size, &USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static int32_t USART1_WriteData (const uint8_t *data, uint32_t size) {
    return UART_WriteData (data, size, &USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static int32_t USART2_WriteData (const uint8_t *data, uint32_t size) {
    return UART_WriteData (data, size, &USART2_Resources);
  }
#endif


/**
  \fn          int32_t UART_ReadData (uint8_t *data,
                                      uint32_t         size,
                                      UART_RESOURCES  *ptr_uart)
  \brief       Read data from UART receiver.
  \param[out]  data  Pointer to buffer for data read from UART receiver
  \param[in]   size  Data buffer size in bytes
  \param[in]   ptr_uart Pointer to UART resources
  \return      number of data bytes read, error code if negative
*/
static int32_t UART_ReadData (uint8_t         *data,
                              uint32_t         size,
                              UART_RESOURCES  *ptr_uart) {
  uint32_t cnt, offs, out;
  int32_t  rd_cnt = 0;
  
  if (ptr_uart->info->power == false) { return rd_cnt; }

  osMutexWait(ptr_uart->info->mutex_read, osWaitForever);

  if (size && !ptr_uart->info->rx.empty) {
    /* Determine number of data bytes in buffer */
    cnt = (ptr_uart->info->rx.idx_in - ptr_uart->info->rx.idx_out) & (ptr_uart->rx_buffer_size - 1);

    if (cnt == 0) {
      /* Rx buffer is full */
      cnt = ptr_uart->rx_buffer_size;
    }

    if (size > cnt) {
      /* Max val of data can be read */
      size = cnt;
    }
    
    offs = ptr_uart->info->rx.idx_out;

    if (size > (ptr_uart->rx_buffer_size - offs)) {
      /* Read data from out to the end of rx buffer */
      rd_cnt = ptr_uart->rx_buffer_size - offs;
      memcpy(data, &ptr_uart->rx_buffer[offs], rd_cnt);
      
      /* Adjust number of data to read and address offset */
      size -= rd_cnt;
      offs  = 0;
    }

    memcpy(&data[rd_cnt], &ptr_uart->rx_buffer[offs], size);
    rd_cnt += size;
    
    /* Precalculate output index to determine if buffer empty */
    out = (ptr_uart->info->rx.idx_out + rd_cnt) & (ptr_uart->rx_buffer_size - 1);

    if (out == ptr_uart->info->rx.idx_in) {
        ptr_uart->info->rx.empty = true;
    }
    ptr_uart->info->rx.idx_out = out;
  }

  /* Enable RX Data Valid interrupt */
  ptr_uart->uart->IEN |= USART_IEN_RXDATAV;

  osMutexRelease(ptr_uart->info->mutex_read);

  return rd_cnt;
}

#if (RTE_UART0)
  static int32_t UART0_ReadData (uint8_t *data, uint32_t size) {
    return UART_ReadData (data, size, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static int32_t UART1_ReadData (uint8_t *data, uint32_t size) {
    return UART_ReadData (data, size, &UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static int32_t USART0_ReadData (uint8_t *data, uint32_t size) {
    return UART_ReadData (data, size, &USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static int32_t USART1_ReadData (uint8_t *data, uint32_t size) {
    return UART_ReadData (data, size, &USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static int32_t USART2_ReadData (uint8_t *data, uint32_t size) {
    return UART_ReadData (data, size, &USART2_Resources);
  }
#endif


/**
  \fn          int32_t UART_DataAvailable (UART_RESOURCES  *ptr_uart)
  \brief       Check available data in UART receiver.
  \param[in]   ptr_uart Pointer to UART resources
  \return      number of data bytes in receive buffer, error code if negative
*/
static int32_t UART_DataAvailable (UART_RESOURCES *ptr_uart) {
  uint32_t cnt;

  cnt = (ptr_uart->info->rx.idx_in - ptr_uart->info->rx.idx_out) & (ptr_uart->rx_buffer_size - 1);

  if ((cnt == 0) && !ptr_uart->info->rx.empty) {
    /* Buffer is full */
    cnt = ptr_uart->rx_buffer_size;
  }
  return (cnt);
}

#if (RTE_UART0)
  static int32_t UART0_DataAvailable (void) {
    return UART_DataAvailable (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static int32_t UART1_DataAvailable (void) {
    return UART_DataAvailable (&UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static int32_t USART0_DataAvailable (void) {
    return UART_DataAvailable (&USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static int32_t USART1_DataAvailable (void) {
    return UART_DataAvailable (&USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static int32_t USART2_DataAvailable (void) {
    return UART_DataAvailable (&USART2_Resources);
  }
#endif


/**
  \fn          bool UART_TxDone (UART_RESOURCES  *ptr_uart)
  \brief       Check if UART transmission is completed.
  \param[in]   ptr_uart Pointer to UART resources
  \return      transmitter completion status
                 - \b true transmission done
                 - \b false transmission in progress
*/
static bool UART_TxDone (UART_RESOURCES *ptr_uart) {
  return (ptr_uart->info->tx.empty);
}

#if (RTE_UART0)
  static bool UART0_TxDone (void) {
    return UART_TxDone (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static bool UART1_TxDone (void) {
    return UART_TxDone (&UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static bool USART0_TxDone (void) {
    return UART_TxDone (&USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static bool USART1_TxDone (void) {
    return UART_TxDone (&USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static bool USART2_TxDone (void) {
    return UART_TxDone (&USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_SetTxThreshold (uint32_t         level,
                                                    UART_RESOURCES  *ptr_uart) {
  \brief       Set transmit threshold for UART_TX_THRESHOLD event.
  \param[in]   level  Number of character in transmit buffer
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_SetTxThreshold (uint32_t level, UART_RESOURCES *ptr_uart) {

  if (level > ptr_uart->tx_buffer_size) { return ARM_UART_ERROR; }
  ptr_uart->info->tx.threshold = level;

  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_SetTxThreshold (uint32_t level) {
    return UART_SetTxThreshold (level, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_SetTxThreshold (uint32_t level) {
    return UART_SetTxThreshold (level, &UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_STATUS USART0_SetTxThreshold (uint32_t level) {
    return UART_SetTxThreshold (level, &USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_STATUS USART1_SetTxThreshold (uint32_t level) {
    return UART_SetTxThreshold (level, &USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_STATUS USART2_SetTxThreshold (uint32_t level) {
    return UART_SetTxThreshold (level, &USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_SetRxThreshold (uint32_t         level,
                                                    UART_RESOURCES  *ptr_uart) {
  \brief       Set receive threshold for UART_RX_THRESHOLD event.
  \param[in]   level  Number of character in transmit buffer
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_SetRxThreshold (uint32_t level, UART_RESOURCES *ptr_uart) {

  if (level > ptr_uart->rx_buffer_size) { return ARM_UART_ERROR; }
  ptr_uart->info->rx.threshold = level;
  
  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_SetRxThreshold (uint32_t level) {
    return UART_SetRxThreshold (level, &UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_SetRxThreshold (uint32_t level) {
    return UART_SetRxThreshold (level, &UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_STATUS USART0_SetRxThreshold (uint32_t level) {
    return UART_SetRxThreshold (level, &USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_STATUS USART1_SetRxThreshold (uint32_t level) {
    return UART_SetRxThreshold (level, &USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_STATUS USART2_SetRxThreshold (uint32_t level) {
    return UART_SetRxThreshold (level, &USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_FlushTxBuffer (UART_RESOURCES  *ptr_uart)
  \brief       Flush UART transmit buffer.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_FlushTxBuffer (UART_RESOURCES *ptr_uart) {

  osMutexWait(ptr_uart->info->mutex_write, osWaitForever);
  
  /* Disable transmitter */
  ptr_uart->uart->CMD  = USART_CMD_TXDIS;
  /* Disable tx buffer empty interrupt */
  ptr_uart->uart->IEN &= ~USART_IEN_TXBL;
  /* Clear tx peripheral buffer and shift register */
  ptr_uart->uart->CMD  = USART_CMD_CLEARTX;

  /* Reset buffer info */
  ptr_uart->info->tx.idx_in  = 0;
  ptr_uart->info->tx.idx_out = 0;
  ptr_uart->info->tx.empty   = true;

  /* Enable transmitter */
  ptr_uart->uart->CMD = USART_CMD_TXEN;

  osMutexRelease(ptr_uart->info->mutex_write);
  
  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_FlushTxBuffer (void) {
    return UART_FlushTxBuffer (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_FlushTxBuffer (void) {
    return UART_FlushTxBuffer (&UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_STATUS USART0_FlushTxBuffer (void) {
    return UART_FlushTxBuffer (&USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_STATUS USART1_FlushTxBuffer (void) {
    return UART_FlushTxBuffer (&USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_STATUS USART2_FlushTxBuffer (void) {
    return UART_FlushTxBuffer (&USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_FlushRxBuffer (UART_RESOURCES  *ptr_uart)
  \brief       Flush UART transmit buffer.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_FlushRxBuffer (UART_RESOURCES *ptr_uart) {

  osMutexWait(ptr_uart->info->mutex_read, osWaitForever);

  /* Disable receiver */
  ptr_uart->uart->CMD  = USART_CMD_RXDIS;
  /* Disable RX data valid interrupt */
  ptr_uart->uart->IEN &= USART_IEN_RXDATAV;
  /* Clear rx peripheral buffer and shift register */
  ptr_uart->uart->CMD  = USART_CMD_CLEARRX;

  /* Reset buffer info */
  ptr_uart->info->rx.idx_in  = 0;
  ptr_uart->info->rx.idx_out = 0;
  ptr_uart->info->rx.empty   = true;

  /* Enable rx data valid interrupt */
  ptr_uart->uart->IEN |= USART_IEN_RXDATAV;
  /* Enable receiver */
  ptr_uart->uart->CMD  = USART_CMD_RXEN;

  osMutexRelease(ptr_uart->info->mutex_read);

  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_FlushRxBuffer (void) {
    return UART_FlushRxBuffer (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_FlushRxBuffer (void) {
    return UART_FlushRxBuffer (&UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_STATUS USART0_FlushRxBuffer (void) {
    return UART_FlushRxBuffer (&USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_STATUS USART1_FlushRxBuffer (void) {
    return UART_FlushRxBuffer (&USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_STATUS USART2_FlushRxBuffer (void) {
    return UART_FlushRxBuffer (&USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_COM_ERROR UART_GetComError (UART_RESOURCES  *ptr_uart)
  \brief       Get UART Communication errors.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_COM_ERROR
*/
static ARM_UART_COM_ERROR UART_GetComError (UART_RESOURCES  *ptr_uart) {
  ARM_UART_COM_ERROR com_error = {0, 0, 0, 0};
  return com_error;
}

#if (RTE_UART0)
  static ARM_UART_COM_ERROR UART0_GetComError (void) {
    return UART_GetComError (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_COM_ERROR UART1_GetComError (void) {
    return UART_GetComError (&UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_COM_ERROR USART0_GetComError (void) {
    return UART_GetComError (&USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_COM_ERROR USART1_GetComError (void) {
    return UART_GetComError (&USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_COM_ERROR USART2_GetComError (void) {
    return UART_GetComError (&USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_SetComBreak (UART_RESOURCES  *ptr_uart)
  \brief       Suspend transmission and put UART Transmission line in a break state.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_SetComBreak (UART_RESOURCES  *ptr_uart) {
  if (ptr_uart->info->power == false) { return ARM_UART_ERROR; }
  
  ptr_uart->info->com_break = true;
  ptr_uart->uart->IEN |= USART_IEN_TXBL;
  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_SetComBreak (void) {
    return UART_SetComBreak (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_SetComBreak (void) {
    return UART_SetComBreak (&UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_STATUS USART0_SetComBreak (void) {
    return UART_SetComBreak (&USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_STATUS USART1_SetComBreak (void) {
    return UART_SetComBreak (&USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_STATUS USART2_SetComBreak (void) {
    return UART_SetComBreak (&USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_ClearComBreak (UART_RESOURCES  *ptr_uart)
  \brief       Suspend transmission and put UART Transmission line in a break state.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref UART_STATUS
*/
static ARM_UART_STATUS UART_ClearComBreak (UART_RESOURCES  *ptr_uart) {
  if (ptr_uart->info->power == false) { return ARM_UART_ERROR; }

  ptr_uart->info->com_break = false;
  ptr_uart->uart->IEN |= USART_IEN_TXBL;
  return ARM_UART_OK;
}

#if (RTE_UART0)
  static ARM_UART_STATUS UART0_ClearComBreak (void) {
    return UART_ClearComBreak (&UART0_Resources);
  }
#endif

#if (RTE_UART1)
  static ARM_UART_STATUS UART1_ClearComBreak (void) {
    return UART_ClearComBreak (&UART1_Resources);
  }
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  static ARM_UART_STATUS USART0_ClearComBreak (void) {
    return UART_ClearComBreak (&USART0_Resources);
  }
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  static ARM_UART_STATUS USART1_ClearComBreak (void) {
    return UART_ClearComBreak (&USART1_Resources);
  }
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  static ARM_UART_STATUS USART2_ClearComBreak (void) {
    return UART_ClearComBreak (&USART2_Resources);
  }
#endif


/**
  \fn          ARM_UART_STATUS UART_SetModemControl (ARM_UART_MODEM_CONTROL control,
                                                     UART_RESOURCES        *ptr_uart)
  \brief       Set UART Modem Control line state.
  \param[in]   control  \ref ARM_UART_MODEM_CONTROL
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_STATUS
*/
static ARM_UART_STATUS UART_SetModemControl (ARM_UART_MODEM_CONTROL control,
                                             UART_RESOURCES        *ptr_uart) {
  return ARM_UART_ERROR;
}

#if (RTE_UART0)
static ARM_UART_STATUS UART0_SetModemControl (ARM_UART_MODEM_CONTROL control) {
  return UART_SetModemControl (control, &UART0_Resources);
}
#endif

#if (RTE_UART1)
static ARM_UART_STATUS UART1_SetModemControl (ARM_UART_MODEM_CONTROL control) {
  return UART_SetModemControl (control, &UART1_Resources);
}
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
static ARM_UART_STATUS USART0_SetModemControl (ARM_UART_MODEM_CONTROL control) {
  return UART_SetModemControl (control, &USART0_Resources);
}
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
static ARM_UART_STATUS USART1_SetModemControl (ARM_UART_MODEM_CONTROL control) {
  return UART_SetModemControl (control, &USART1_Resources);
}
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
static ARM_UART_STATUS USART2_SetModemControl (ARM_UART_MODEM_CONTROL control) {
  return UART_SetModemControl (control, &USART2_Resources);
}
#endif


/**
  \fn          ARM_UART_MODEM_STATUS UART_GetModemStatus (UART_RESOURCES *ptr_uart)
  \brief       Get UART Modem Status lines state.
  \param[in]   ptr_uart Pointer to UART resources
  \return      \ref ARM_UART_MODEM_STATUS
*/
static ARM_UART_MODEM_STATUS UART_GetModemStatus (UART_RESOURCES *ptr_uart) {
  ARM_UART_MODEM_STATUS mst = {0, 0, 0, 0};
  return mst;
}

#if (RTE_UART0)
static ARM_UART_MODEM_STATUS UART0_GetModemStatus (void) {
  return UART_GetModemStatus (&UART0_Resources);
}
#endif

#if (RTE_UART1)
static ARM_UART_MODEM_STATUS UART1_GetModemStatus (void) {
  return UART_GetModemStatus (&UART1_Resources);
}
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
static ARM_UART_MODEM_STATUS USART0_GetModemStatus (void) {
  return UART_GetModemStatus (&USART0_Resources);
}
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
static ARM_UART_MODEM_STATUS USART1_GetModemStatus (void) {
  return UART_GetModemStatus (&USART1_Resources);
}
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
static ARM_UART_MODEM_STATUS USART2_GetModemStatus (void) {
  return UART_GetModemStatus (&USART2_Resources);
}
#endif


/**
  \fn          void UART_Rx_IRQHandler (UART_RESOURCES *ptr_uart)
  \brief       UART receive interrupt handler.
  \param[in]   ptr_uart  Pointer to UART resources
*/
static void UART_Rx_IRQHandler (UART_RESOURCES *ptr_uart) {
  uint32_t cnt;
  uint32_t flags = ptr_uart->uart->IF;

  if (flags & USART_IF_RXDATAV) {
    /* Received data valid interrupt */
    cnt = (ptr_uart->info->rx.idx_in - ptr_uart->info->rx.idx_out) & (ptr_uart->rx_buffer_size - 1);
    
    if (cnt || ptr_uart->info->rx.empty) {
      /* Free space in buffer or buffer empty */
      ptr_uart->rx_buffer[ptr_uart->info->rx.idx_in] = ptr_uart->uart->RXDATA;

      ptr_uart->info->rx.idx_in += 1;
      ptr_uart->info->rx.idx_in &= ptr_uart->rx_buffer_size - 1;
      
      if (ptr_uart->info->rx.empty) {
        if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_NOT_EMPTY)) {
          ptr_uart->info->cb_event(ARM_UART_EVENT_RX_NOT_EMPTY);
        }
      }
      ptr_uart->info->rx.empty = false;

      cnt += 1;

      if (cnt == ptr_uart->info->rx.threshold) {
        if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_THRESHOLD)) {
          ptr_uart->info->cb_event(ARM_UART_EVENT_RX_THRESHOLD);
        }
      }
    }
    else {
      /* Rx buffer full, disable data available interrupt */
      ptr_uart->uart->IEN &= ~USART_IEN_RXDATAV;
    }
  }
  else if (flags & USART_IF_FERR) {
    /* Received frame with framing error */
    ptr_uart->uart->IFC = USART_IFC_FERR;

    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_FRAMING_ERROR)) {
      ptr_uart->info->cb_event(ARM_UART_EVENT_RX_FRAMING_ERROR);
    }
  }
  else if (flags & USART_IF_PERR) {
    /* Received frame with a parity error */
    ptr_uart->uart->IFC = USART_IFC_PERR;

    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_PARITY_ERROR)) {
      ptr_uart->info->cb_event(ARM_UART_EVENT_RX_PARITY_ERROR);
    }
  }
  else if (flags & USART_IF_RXOF) {
    /* Rx overflow: incoming data while rx shift register is full */
    ptr_uart->uart->IFC = USART_IFC_RXOF;

    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_RX_OVERRUN)) {
      ptr_uart->info->cb_event(ARM_UART_EVENT_RX_OVERRUN);
    }
  }
}


/**
  \fn          void UART_Tx_IRQHandler (UART_RESOURCES *ptr_uart)
  \brief       UART transmit interrupt handler.
  \param[in]   ptr_uart  Pointer to UART resources
*/
static void UART_Tx_IRQHandler (UART_RESOURCES *ptr_uart) {
  uint32_t cnt;

  if (ptr_uart->uart->IF & USART_IF_TXBL) {
    /* Peripheral tx buffer empty interrupt */
    if (ptr_uart->info->tx.empty) {
      if (ptr_uart->info->com_break) {
        /* Send break condition */
        ptr_uart->uart->TXDATAX = USART_TXDATAX_TXBREAK;
      }
      /* Disable TX empty interrupt */
      ptr_uart->uart->IEN &= ~USART_IEN_TXBL;
    }
    else {
      /* Data to transmit exists in tx buffer */
      ptr_uart->uart->TXDATA = ptr_uart->tx_buffer[ptr_uart->info->tx.idx_out];

      ptr_uart->info->tx.idx_out += 1;
      ptr_uart->info->tx.idx_out &= ptr_uart->tx_buffer_size - 1;

      /* Determine how may data is in the buffer */
      cnt = (ptr_uart->info->tx.idx_in - ptr_uart->info->tx.idx_out) & (ptr_uart->tx_buffer_size - 1);

      if ((ptr_uart->tx_buffer_size - cnt) == ptr_uart->info->tx.threshold) {
        /* Free space in buffer reached the tx threshold */
        if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_THRESHOLD)) {
          ptr_uart->info->cb_event(ARM_UART_EVENT_TX_THRESHOLD);
        }
      }

      if (cnt == 0) {
        /* All data from tx buffer sent - buffer is empty */
        ptr_uart->info->tx.empty = true;

        if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_EMPTY)) {
          ptr_uart->info->cb_event(ARM_UART_EVENT_TX_EMPTY);
        }
      }
    }
  }

  if (ptr_uart->uart->IF & USART_IF_TXC) {
    /* Transmition complete (tx and shift registers empty) */
    ptr_uart->uart->IFC = USART_IFC_TXC;

    if (ptr_uart->info->tx.empty) {
      if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_DONE)) {
        ptr_uart->info->cb_event(ARM_UART_EVENT_TX_DONE);
      }
    }
  }
}

#if (RTE_UART0)
void UART0_RX_IRQHandler (void) {
  UART_Rx_IRQHandler(&UART0_Resources);
}
void UART0_TX_IRQHandler (void) {
  UART_Tx_IRQHandler(&UART0_Resources);
}
#endif

#if (RTE_UART1)
void UART1_RX_IRQHandler (void) {
  UART_Rx_IRQHandler(&UART1_Resources);
}
void UART1_TX_IRQHandler (void) {
  UART_Tx_IRQHandler(&UART1_Resources);
}
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
void USART0_RX_IRQHandler (void) {
  UART_Rx_IRQHandler(&USART0_Resources);
}
void USART0_TX_IRQHandler (void) {
  UART_Tx_IRQHandler(&USART0_Resources);
}
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
void USART1_RX_IRQHandler (void) {
  UART_Rx_IRQHandler(&USART1_Resources);
}
void USART1_TX_IRQHandler (void) {
  UART_Tx_IRQHandler(&USART1_Resources);
}
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
void USART2_RX_IRQHandler (void) {
  UART_Rx_IRQHandler(&USART2_Resources);
}
void USART2_TX_IRQHandler (void) {
  UART_Tx_IRQHandler(&USART2_Resources);
}
#endif


/* DMA Callbacks */

#if defined(UART_TX_DMA_ENABLE)
/**
  \fn          void USART_DMA_TxComplete (USART_RESOURCES *ptr_usart)
  \brief       DMA transmit complete callback.
  \param[in]   ptr_usart  Pointer to USART resources
*/
static void UART_DMA_TxComplete (UART_RESOURCES *ptr_uart) {
  DMA_CfgDescr_TypeDef DmaDescCfg;
  uint32_t cnt, free;
  
  if (ptr_uart->info->dma_xfer_len) {
    /* DMA completed the data transfer */
    ptr_uart->info->tx.idx_out += ptr_uart->info->dma_xfer_len;
    ptr_uart->info->tx.idx_out &= ptr_uart->tx_buffer_size - 1;
  }

  /* Determine how many data is in the tx buffer */
  cnt = (ptr_uart->info->tx.idx_in - ptr_uart->info->tx.idx_out) & (ptr_uart->tx_buffer_size - 1);

  /* Get free space in buffer */
  free = ptr_uart->tx_buffer_size - cnt;

  if (free == ptr_uart->info->tx.threshold) {
    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_THRESHOLD)) {
      ptr_uart->info->cb_event(ARM_UART_EVENT_TX_THRESHOLD);
    }
  }
  
  if (cnt == 0) {
    /* Tx buffer is empty */
    ptr_uart->info->tx.empty = true;

    if (ptr_uart->info->event_mask & (1 << ARM_UART_EVENT_TX_EMPTY)) {
      ptr_uart->info->cb_event(ARM_UART_EVENT_TX_EMPTY);
    }
    osSemaphoreRelease(ptr_uart->info->semaphore);
  }
  else {
    /* Determine how many data can be sent till next limit */
    if (cnt > free) {
      /* Tx threshold event limit */
      cnt = free;
    }
    if (cnt > (ptr_uart->tx_buffer_size - ptr_uart->info->tx.idx_out)) {
      /* End of buffer limit */
      cnt = ptr_uart->tx_buffer_size - ptr_uart->info->tx.idx_out;
    }
    
    /* Remember current DMA transfer length */
    ptr_uart->info->dma_xfer_len = cnt;

    /* Configure DMA descriptor to send "limit" of data until next interrupt */
    DmaDescCfg.dstInc  = dmaDataIncNone;
    DmaDescCfg.srcInc  = dmaDataInc1;
    DmaDescCfg.size    = dmaDataSize1;
    DmaDescCfg.arbRate = dmaArbitrate1;
    DmaDescCfg.hprot   = 0;

    DMA_CfgDescr (ptr_uart->dma_tx_ch, true, &DmaDescCfg);

    DMA_ActivateBasic (ptr_uart->dma_tx_ch, true, false,
                       (void *)&ptr_uart->uart->TXDATA,
                       (void *)(ptr_uart->tx_buffer + ptr_uart->info->tx.idx_out),
                       cnt-1);
  }
}
#endif /* UART_TX_DMA_ENABLE */

void UART_DMA_Complete (uint32_t channel, bool primary, void *userPtr) {
  #if (RTE_UART0 && RTE_UART0_TX_DMA)
  if (channel == RTE_UART0_TX_DMA_CHANNEL) {
    UART_DMA_TxComplete (&UART0_Resources);
  }
  #endif

  #if (RTE_UART1 && RTE_UART1_TX_DMA)
  if (channel == RTE_UART1_TX_DMA_CHANNEL) {
    UART_DMA_TxComplete (&UART1_Resources);
  }
  #endif

  #if (RTE_USART0 && (RTE_USART0_SYNC == 0) && RTE_USART0_TX_DMA)
  if (channel == RTE_USART0_TX_DMA_CHANNEL) {
    UART_DMA_TxComplete (&USART0_Resources);
  }
  #endif
  
  #if (RTE_USART1 && (RTE_USART1_SYNC == 0) && RTE_USART1_TX_DMA)
  if (channel == RTE_USART1_TX_DMA_CHANNEL) {
    UART_DMA_TxComplete (&USART1_Resources);
  }
  #endif

  #if (RTE_USART2 && (RTE_USART2_SYNC == 0) && RTE_USART2_TX_DMA)
  if (channel == RTE_USART2_TX_DMA_CHANNEL) {
    UART_DMA_TxComplete (&USART2_Resources);
  }
  #endif
}

#if (RTE_UART0)
  /* UART0 peripheral exports Driver_UART0 */
  ARM_DRIVER_UART Driver_UART0 = {
    UARTX_GetVersion,
    UART0_GetCapabilities,
    UART0_Initialize,
    UART0_Uninitialize,
    UART0_PowerControl,
    UART0_Configure,
    UART0_WriteData,
    UART0_ReadData,
    UART0_DataAvailable,
    UART0_TxDone,
    UART0_SetTxThreshold,
    UART0_SetRxThreshold,
    UART0_FlushTxBuffer,
    UART0_FlushRxBuffer,
    UART0_GetComError,
    UART0_SetComBreak,
    UART0_ClearComBreak,
    UART0_SetModemControl,
    UART0_GetModemStatus
  };
#endif

#if (RTE_UART1)
  /* UART1 peripheral exports Driver_UART1 */
  ARM_DRIVER_UART Driver_UART1 = {
    UARTX_GetVersion,
    UART1_GetCapabilities,
    UART1_Initialize,
    UART1_Uninitialize,
    UART1_PowerControl,
    UART1_Configure,
    UART1_WriteData,
    UART1_ReadData,
    UART1_DataAvailable,
    UART1_TxDone,
    UART1_SetTxThreshold,
    UART1_SetRxThreshold,
    UART1_FlushTxBuffer,
    UART1_FlushRxBuffer,
    UART1_GetComError,
    UART1_SetComBreak,
    UART1_ClearComBreak,
    UART1_SetModemControl,
    UART1_GetModemStatus
  };
#endif

#if (RTE_USART0 && (RTE_USART0_SYNC == 0))
  /* USART0 peripheral exports Driver_UART2 */
  ARM_DRIVER_UART Driver_UART2 = {
    UARTX_GetVersion,
    USART0_GetCapabilities,
    USART0_Initialize,
    USART0_Uninitialize,
    USART0_PowerControl,
    USART0_Configure,
    USART0_WriteData,
    USART0_ReadData,
    USART0_DataAvailable,
    USART0_TxDone,
    USART0_SetTxThreshold,
    USART0_SetRxThreshold,
    USART0_FlushTxBuffer,
    USART0_FlushRxBuffer,
    USART0_GetComError,
    USART0_SetComBreak,
    USART0_ClearComBreak,
    USART0_SetModemControl,
    USART0_GetModemStatus
  };
#endif

#if (RTE_USART1 && (RTE_USART1_SYNC == 0))
  /* USART1 peripheral exports Driver_UART3 */
  ARM_DRIVER_UART Driver_UART3 = {
    UARTX_GetVersion,
    USART1_GetCapabilities,
    USART1_Initialize,
    USART1_Uninitialize,
    USART1_PowerControl,
    USART1_Configure,
    USART1_WriteData,
    USART1_ReadData,
    USART1_DataAvailable,
    USART1_TxDone,
    USART1_SetTxThreshold,
    USART1_SetRxThreshold,
    USART1_FlushTxBuffer,
    USART1_FlushRxBuffer,
    USART1_GetComError,
    USART1_SetComBreak,
    USART1_ClearComBreak,
    USART1_SetModemControl,
    USART1_GetModemStatus
  };
#endif

#if (RTE_USART2 && (RTE_USART2_SYNC == 0))
  /* USART2 peripheral exports Driver_UART4 */
  ARM_DRIVER_UART Driver_UART4 = {
    UARTX_GetVersion,
    USART2_GetCapabilities,
    USART2_Initialize,
    USART2_Uninitialize,
    USART2_PowerControl,
    USART2_Configure,
    USART2_WriteData,
    USART2_ReadData,
    USART2_DataAvailable,
    USART2_TxDone,
    USART2_SetTxThreshold,
    USART2_SetRxThreshold,
    USART2_FlushTxBuffer,
    USART2_FlushRxBuffer,
    USART2_GetComError,
    USART2_SetComBreak,
    USART2_ClearComBreak,
    USART2_SetModemControl,
    USART2_GetModemStatus
  };
#endif
