/* ----------------------------------------------------------------------
 * Copyright (C) 2013 ARM Limited. All rights reserved.
 *  
 * $Date:        30. October 2013
 * $Revision:    V1.04
 *  
 * Driver:       Driver_USBD0
 * Configured:   via RTE_Device.h configuration file 
 * Project:      USB Full/Low-Speed Device Driver
 *               for EnergyMicro EFM32GGxxx
 * ---------------------------------------------------------------------- 
 * Use the following configuration settings in the middleware component
 * to connect to this driver.
 * 
 *   Configuration Setting                Value
 *   ---------------------                -----
 *   Connect to hardware via Driver_USBD# = 0
 * -------------------------------------------------------------------- */

/* History:
 *  Version 1.04
 *    Multiple packet read
 *  Version 1.03
 *    Based on API V1.10 (namespace prefix ARM_ added)
 *  Version 1.00
 *    Initial release
 */

#include <string.h>
#include "cmsis_os.h"

#include "Driver_USBD.h"

#include "em_device.h"

#include "em_cmu.h"
#include "em_emu.h"
#include "em_gpio.h"

#include "RTE_Device.h"
#include "RTE_Components.h"

extern uint8_t otg_role;

extern bool OTG_PinsConfigure   (uint8_t pins_mask);
extern bool OTG_PinsUnconfigure (uint8_t pins_mask);
extern bool OTG_PinVbusOnOff    (bool state);


/* USBD Driver ****************************************************************/

#define ARM_USBD_DRV_VERSION ARM_DRV_VERSION_MAJOR_MINOR(1,04) /* USBD driver version */

/* Driver Version */
static const ARM_DRV_VERSION usbd_driver_version = { ARM_USBD_API_VERSION, ARM_USBD_DRV_VERSION };

/* Driver Capabilities */
static const ARM_USBD_CAPABILITIES usbd_driver_capabilities = {
  false,  /* event_power_on      */
  false,  /* event_power_off     */
  true,   /* event_connect       */
  false,  /* event_disconnect    */
  true,   /* event_reset         */
  true,   /* event_high_speed    */
  true,   /* event_suspend       */
  true,   /* event_resume        */
  false,  /* event_remote_wakeup */
  false   /* reserved            */
};

// Number of endpoints
#if   (RTE_USB_DEV_EP & 0x40)
#define USBD_EP_NUM      6
#elif (RTE_USB_DEV_EP & 0x20)
#define USBD_EP_NUM      5
#elif (RTE_USB_DEV_EP & 0x10)
#define USBD_EP_NUM      4
#elif (RTE_USB_DEV_EP & 0x08)
#define USBD_EP_NUM      3
#elif (RTE_USB_DEV_EP & 0x04)
#define USBD_EP_NUM      2
#elif (RTE_USB_DEV_EP & 0x02)
#define USBD_EP_NUM      1
#else
#define USBD_EP_NUM      0
#endif

// Interrupt IN Endpoint intermediate buffers
#define USB_INT_IN_EP        ((RTE_USB_DEV_EP_INT >> 16) & RTE_USB_DEV_EP)
#if (USB_INT_IN_EP != 0)

#define USB_INT_IN_EP1       ((USB_INT_IN_EP >> 1) & 1)
#define USB_INT_IN_EP2       ((USB_INT_IN_EP >> 2) & 1)
#define USB_INT_IN_EP3       ((USB_INT_IN_EP >> 3) & 1)
#define USB_INT_IN_EP4       ((USB_INT_IN_EP >> 4) & 1)
#define USB_INT_IN_EP5       ((USB_INT_IN_EP >> 5) & 1)
#define USB_INT_IN_EP6       ((USB_INT_IN_EP >> 6) & 1)

#define USB_INT_IN_EP_BUF_SZ  (USB_INT_IN_EP1 * 64 +         \
                               USB_INT_IN_EP2 * 64 +         \
                               USB_INT_IN_EP3 * 64 +         \
                               USB_INT_IN_EP4 * 64 +         \
                               USB_INT_IN_EP5 * 64 +         \
                               USB_INT_IN_EP6 * 64 )
                             
      uint8_t   USB_IntInEpTmpBuf[USB_INT_IN_EP_BUF_SZ];
const uint8_t  *ptr_USB_IntInEpTmpBuf[] = {
  0,
  USB_INT_IN_EP1 ?  USB_IntInEpTmpBuf                                                                                               : 0,
  USB_INT_IN_EP2 ? (USB_IntInEpTmpBuf +  (USB_INT_IN_EP1 * 64)                 )                                                    : 0,
  USB_INT_IN_EP3 ? (USB_IntInEpTmpBuf + ((USB_INT_IN_EP1 + USB_INT_IN_EP2)* 64))                                                    : 0,
  USB_INT_IN_EP4 ? (USB_IntInEpTmpBuf + ((USB_INT_IN_EP1 + USB_INT_IN_EP2 + USB_INT_IN_EP3)* 64))                                   : 0,
  USB_INT_IN_EP5 ? (USB_IntInEpTmpBuf + ((USB_INT_IN_EP1 + USB_INT_IN_EP2 + USB_INT_IN_EP3 + USB_INT_IN_EP4)* 64))                  : 0,
  USB_INT_IN_EP6 ? (USB_IntInEpTmpBuf + ((USB_INT_IN_EP1 + USB_INT_IN_EP2 + USB_INT_IN_EP3 + USB_INT_IN_EP4 + USB_INT_IN_EP5)* 64)) : 0
};
#endif

#define USB_RX_FIFO_SIZE        256
#define USB_TX0_FIFO_SIZE       64
#define USB_TX1_FIFO_SIZE       64
#define USB_TX2_FIFO_SIZE       64
#define USB_TX3_FIFO_SIZE       64
#define USB_TX4_FIFO_SIZE       64
#define USB_TX5_FIFO_SIZE       64
#define USB_TX6_FIFO_SIZE       64

#define USB_TX_FIFO(n)          *((__packed volatile uint32_t*)(USB->FIFO0D + n*1024))

#define USB_DIEPTSIZ(EPNum)     *(&USB->DIEP0TSIZ    + EPNum * 8)
#define USB_DIEPCTL(EPNum)      *(&USB->DIEP0CTL     + EPNum * 8)
#define USB_DTXFSTS(EPNum)      *(&USB->DIEP0TXFSTS  + EPNum * 8)
#define USB_DOEPTSIZ(EPNum)     *(&USB->DOEP0TSIZ    + EPNum * 8)
#define USB_DOEPCTL(EPNum)      *(&USB->DOEP0CTL     + EPNum * 8)
#define USB_DIEPINT(EPNum)      *(&USB->DIEP0INT     + EPNum * 8)
#define USB_DOEPINT(EPNum)      *(&USB->DOEP0INT     + EPNum * 8)

#define USB_EP_IN_TYPE(num)      ((USB_DIEPCTL(num) >> 18) & 3)
#define USB_EP_OUT_TYPE(num)     ((USB_DOEPCTL(num) >> 18) & 3)


/* Endpoint structure */
typedef struct {
  uint8_t  *buffer;
  uint32_t  bufferIndex;
  uint32_t  dataSize;
  uint32_t  maxPacketSize;
  bool      startRequired;
} ENDPOINT;

/* Static Variables */
static ARM_USBD_SignalDeviceEvent_t   cbDeviceEvent;
static ARM_USBD_SignalEndpointEvent_t cbEndpointEvent;

static ARM_USBD_STATE UsbdState = {0, 0, 0, 0};

volatile static ENDPOINT OutEndpoint[7];
volatile static ENDPOINT InEndpoint[7];
volatile static uint32_t InEndpointNAK;


/* LOCAL FUNCTIONS */

/**
  \fn          bool USBD_FlushInEpFifo (uint8_t ep_addr)
  \brief       Flush IN endpoint fifo
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
*/
static void USBD_FlushInEpFifo (uint8_t ep_addr) {
  ep_addr &= ARM_USB_ENDPOINT_NUMBER_MASK;

  // Flush transmit FIFO 
  USB->GRSTCTL = (USB->GRSTCTL & ~_USB_GRSTCTL_TXFNUM_MASK) |
                 (ep_addr << _USB_GRSTCTL_TXFNUM_SHIFT )    |
                  USB_GRSTCTL_TXFFLSH;
}

/**
  \fn          void USBD_Reset (uint8_t ep_addr)
  \brief       Called after usbd reset interrupt to reset configuration
*/
static void USBD_Reset (void) {
  uint32_t i;

  // Clear enpoint mask registers
  USB->DOEPMSK      = 0;
  USB->DIEPMSK      = 0;  

  for (i = 1; i <= USBD_EP_NUM; i++) {
    USBD_FlushInEpFifo(i);
    if (USB_DOEPCTL(i) & USB_DOEP_CTL_EPENA) {
      USB_DOEPCTL(i)   = USB_DOEP_CTL_EPDIS | // Endpoint disable
                         USB_DOEP_CTL_SNAK;   // Endpoint set NAK
    }
    if (USB_DIEPCTL(i) & USB_DIEP_CTL_EPENA) {
      USB_DIEPCTL(i)   = USB_DIEP_CTL_EPDIS | // Endpoint disable
                         USB_DIEP_CTL_SNAK;   // Endpoint set NAK
    }

    // Reset endpoint resources
    OutEndpoint[i].buffer         = NULL;
    OutEndpoint[i].dataSize       = 0;
    OutEndpoint[i].maxPacketSize  = 0;
    OutEndpoint[i].startRequired  = 0;
    InEndpoint[i].buffer          = NULL;
    InEndpoint[i].dataSize        = 0;
    InEndpoint[i].maxPacketSize   = 0;
    InEndpoint[i].startRequired   = 0;
    
    // Clear IN endpoint interrupts
    USB_DIEPINT(i) = USB_DIEP_INT_XFERCOMPL   |
                     USB_DIEP_INT_EPDISBLD    |
                     USB_DIEP_INT_INTKNTXFEMP |
                     USB_DIEP_INT_INEPNAKEFF  |
                     USB_DIEP_INT_TXFEMP;

    // Clear OUT endpoint interrupts
    USB_DOEPINT(i) = USB_DOEP_INT_XFERCOMPL   |
                     USB_DOEP_INT_EPDISBLD    |
                     USB_DOEP_INT_SETUP       |
                     USB_DOEP_INT_OUTTKNEPDIS;
  }

  USB->DAINTMSK  = USB_DAINTMSK_INEPMSK0 |  // Enable IN endpoint0 interrupt
                   USB_DAINTMSK_OUTEPMSK0;  // Enable OUT endpoint0 interrupt

  // Enable Setup phase done, Out endpoint disabled and Out transfer
  // complete interrupt
  USB->DOEPMSK   = USB_DOEPMSK_SETUPMSK      |
                   USB_DOEPMSK_EPDISBLDMSK   |
                   USB_DOEPMSK_XFERCOMPLMSK;
                      
  // Enable In endpoint disable and In transfer complete interrupt
  USB->DIEPMSK   = USB_DIEPMSK_EPDISBLDMSK     |
                   USB_DIEPMSK_XFERCOMPLMSK;

  // Set device address to 0
  USB->DCFG = (USB->DCFG & ~_USB_DCFG_DEVADDR_MASK) | (0 << _USB_DCFG_DEVADDR_SHIFT);

  // Configure FIFOs
  USB->GRXFSIZ   =  USB_RX_FIFO_SIZE   / 4;
  USB->GNPTXFSIZ = (USB_RX_FIFO_SIZE   / 4) |
                   ((USB_TX0_FIFO_SIZE / 4) << 16);

  USB->DIEPTXF1  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE) / 4) | ((USB_TX1_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF2  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE) / 4) | ((USB_TX2_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF3  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE) / 4) | ((USB_TX3_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF3  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE + USB_TX4_FIFO_SIZE) / 4) | ((USB_TX5_FIFO_SIZE / 4) << 16);
  USB->DIEPTXF3  = ((USB_RX_FIFO_SIZE + USB_TX0_FIFO_SIZE + USB_TX1_FIFO_SIZE + USB_TX2_FIFO_SIZE + USB_TX3_FIFO_SIZE + USB_TX4_FIFO_SIZE + USB_TX5_FIFO_SIZE) / 4) | ((USB_TX6_FIFO_SIZE / 4) << 16);
  

  // Configure Endpoint 0
  InEndpoint[0].maxPacketSize  = 64;
  OutEndpoint[0].maxPacketSize = 64;
  
  USB_DIEPCTL(0)  = USB_DIEP_CTL_SNAK;                // Set NAK
  USB_DOEPCTL(0)  = USB_DOEP_CTL_SNAK;                // Set NAK
  USB_DOEPTSIZ(0) = (3 << _USB_DIEP_TSIZ_MC_SHIFT);   // Setup Packet Count

}


/* FUNCTION PROTOTYPES */
uint16_t GetFrameNumber (void);


/* USB DEVICE DRIVER FUNCTIONS */

/**
  \fn          ARM_DRV_VERSION USBD_GetVersion (void)
  \brief       Get driver version.
  \return      \ref ARM_DRV_VERSION
*/
static ARM_DRV_VERSION GetVersion (void) {
  return usbd_driver_version;
}

/**
  \fn          ARM_USBD_CAPABILITIES USBD_GetCapabilities (void)
  \brief       Get driver capabilities.
  \return      \ref ARM_USBD_CAPABILITIES
*/
static ARM_USBD_CAPABILITIES GetCapabilities (void) {
  return usbd_driver_capabilities;
}

/**
  \fn          ARM_USBD_STATUS Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                           ARM_USBD_SignalEndpointEvent_t cb_endpoint_event)
  \brief       Initialize USB Device Interface.
  \param[in]   cb_device_event    Pointer to \ref ARM_USBD_SignalDeviceEvent
  \param[in]   cb_endpoint_event  Pointer to \ref ARM_USBD_SignalEndpointEvent
  \return      \ref USBD_STATUS
*/
static ARM_USBD_STATUS Initialize (ARM_USBD_SignalDeviceEvent_t   cb_device_event,
                                   ARM_USBD_SignalEndpointEvent_t cb_endpoint_event) {

  cbDeviceEvent   = cb_device_event;
  cbEndpointEvent = cb_endpoint_event;

#if (!RTE_USB_OTG)
  CMU->HFCORECLKEN0 |= CMU_HFCORECLKEN0_USBC;
  CMU_ClockSelectSet( cmuClock_USBC, cmuSelect_HFCLK );

  OTG_PinsConfigure(ARM_USB_PIN_DP | ARM_USB_PIN_DM);
#endif

  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS Uninitialize (void)
  \brief       De-initialize USB Device Interface.
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS Uninitialize (void) {

#if (!RTE_USB_OTG)
  // Enable USB clock
  CMU->HFCORECLKEN0 |= CMU_HFCORECLKEN0_USB;

  NVIC_DisableIRQ(USB_IRQn);

  OTG_PinsUnconfigure(ARM_USB_PIN_DP | ARM_USB_PIN_DM);

  // Disable USB clock
  CMU->HFCORECLKEN0 &= ~(CMU_HFCORECLKEN0_USB | CMU_HFCORECLKEN0_USBC);
#endif
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS PowerControl (ARM_POWER_STATE state)
  \brief       Control USB Device Interface Power.
  \param[in]   state specifies Power state
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS PowerControl (ARM_POWER_STATE state) {
  uint32_t tick;

  switch (state) {
    case ARM_POWER_OFF:
      NVIC_DisableIRQ(USB_IRQn);                    // Disable USB interrupt

      CMU->HFCORECLKEN0 |=  CMU_HFCORECLKEN0_USB;   // Enable USB clock
      USB->DCTL         |=  USB_DCTL_SFTDISCON;     // Soft disconnect enabled
      USB->GAHBCFG      &= ~USB_GAHBCFG_GLBLINTRMSK;// Disable global interrupt mask

      otg_role = ARM_USB_ROLE_NONE;

      // Core soft reset
      USB->GRSTCTL      |=  USB_GRSTCTL_CSFTRST;
      tick = osKernelSysTick();
      do {
        if ((USB->GRSTCTL & USB_GRSTCTL_CSFTRST) == 0) break;
      } while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(10000));
      if (USB->GRSTCTL & USB_GRSTCTL_CSFTRST) return ARM_USBD_ERROR;

      // Disable USB clock
      CMU->HFCORECLKEN0 &= ~(CMU_HFCORECLKEN0_USB | CMU_HFCORECLKEN0_USBC);
      break;

    case ARM_POWER_LOW:
      return ARM_USBD_ERROR;

    case ARM_POWER_FULL:
      CMU->HFCORECLKEN0 |=  CMU_HFCORECLKEN0_USB;   // Enable USB clock

      // Wait until AHB Master state machine is in the idle condition
      tick = osKernelSysTick();
      do {
        if (USB->GRSTCTL & USB_GRSTCTL_AHBIDLE) break;
      } while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(100000));
      
      if ((USB->GRSTCTL & USB_GRSTCTL_AHBIDLE) == 0) return ARM_USBD_ERROR;

      // Core soft reset
      USB->GRSTCTL      |=  USB_GRSTCTL_CSFTRST;
      tick = osKernelSysTick();
      do {
        if ((USB->GRSTCTL & USB_GRSTCTL_CSFTRST) == 0) break;
      } while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(10000));
      if (USB->GRSTCTL & USB_GRSTCTL_CSFTRST) return ARM_USBD_ERROR;

      osDelay(3);

      USB->GAHBCFG      &= ~USB_GAHBCFG_GLBLINTRMSK;// Disable global interrupt mask
      USB->DCTL         |=  USB_DCTL_SFTDISCON;     // Soft disconnect enabled

      // Set turnaround time and force device mode
      USB->GUSBCFG   =  (USB->GUSBCFG & ~_USB_GUSBCFG_USBTRDTIM_MASK) |
                        (5 << _USB_GUSBCFG_USBTRDTIM_SHIFT)           |
                         USB_GUSBCFG_FORCEDEVMODE;
      osDelay(100);

      otg_role = ARM_USB_ROLE_DEVICE;

      // Clear interrupts
      USB->GINTSTS   =   USB_GINTSTS_CURMOD       |
                         USB_GINTSTS_SOF          |
                         USB_GINTSTS_ERLYSUSP     |
                         USB_GINTSTS_USBSUSP      |
                         USB_GINTSTS_USBRST       |
                         USB_GINTSTS_ENUMDONE     |
                         USB_GINTSTS_ISOOUTDROP   |
                         USB_GINTSTS_INCOMPISOIN  |
                         USB_GINTSTS_INCOMPLP     |
                         USB_GINTSTS_FETSUSP      |
                         USB_GINTSTS_RESETDET     |
                         USB_GINTSTS_CONIDSTSCHNG |
                         USB_GINTSTS_DISCONNINT   |
                         USB_GINTSTS_SESSREQINT   |
                         USB_GINTSTS_WKUPINT;

      USB->GINTMSK   =   USB_GINTSTS_USBSUSP      |
                         USB_GINTSTS_USBRST       |
                         USB_GINTSTS_ENUMDONE     |
                         USB_GINTSTS_RXFLVL       |
                         USB_GINTSTS_IEPINT       |
                         USB_GINTSTS_OEPINT       |
                         USB_GINTSTS_WKUPINT;

      // Enable interrupts
      NVIC_EnableIRQ(USB_IRQn);
      USB->GAHBCFG  |=   USB_GAHBCFG_GLBLINTRMSK |
                         USB_GAHBCFG_NPTXFEMPLVL;

      break;
  }

  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS DeviceConnect (void)
  \brief       Connect USB Device.
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS DeviceConnect (void) {
  USB->DCTL &= ~USB_DCTL_SFTDISCON;     // Soft disconnect disabled

  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS DeviceDisconnect (void)
  \brief       Disconnect USB Device.
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS DeviceDisconnect (void) {
  USB->DCTL |= USB_DCTL_SFTDISCON;      // Soft disconnect enabled

  UsbdState.connected = false;
  UsbdState.active    = false;
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATE DeviceGetState (void)
  \brief       Get current USB Device State.
  \return      \ref ARM_USBD_STATE
*/
static ARM_USBD_STATE DeviceGetState (void) {
  return UsbdState;
}

/**
  \fn          ARM_USBD_STATUS DeviceRemoteWakeup (void)
  \brief       Trigger USB Remote Wakeup.
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS DeviceRemoteWakeup (void) {
  USB->DCTL   |=   USB_DCTL_RMTWKUPSIG;       // Remote wakeup signalling
  osDelay(5);
  USB->DCTL   &=  ~USB_DCTL_RMTWKUPSIG;
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS DeviceSetAddress (uint8_t dev_addr,
                                                 ARM_USBD_SET_ADDRESS_STAGE stage)
  \brief       Set USB Device Address.
  \param[in]   dev_addr specifies Device Address
  \param[in]   stage specifies stage in which the function is called
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS DeviceSetAddress (uint8_t dev_addr,
                                         ARM_USBD_SET_ADDRESS_STAGE stage) {
  USB->DCFG = (USB->DCFG & ~_USB_DCFG_DEVADDR_MASK) | 
              (dev_addr << _USB_DCFG_DEVADDR_SHIFT);

  return ARM_USBD_OK;
}

  /**
  \fn          ARM_USBD_STATUS DeviceConfigure (bool configure)
  \brief       Configure/unconfigure USB Device.
  \param[in]   configure specifies operation
                - \b false Unconfigure
                - \b true  Configure
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS DeviceConfigure (bool configure) {
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS USBD_EndpointConfigure (uint8_t ep_addr,
                                                       ARM_USB_ENDPOINT_TYPE ep_type,
                                                       uint16_t ep_max_packet_size)
  \brief       Configure USB Endpoint.
  \param[in]   ptr_epd specifies pointer to Endpoint descriptor
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS EndpointConfigure (uint8_t ep_addr,
                                          ARM_USB_ENDPOINT_TYPE ep_type,
                                          uint16_t ep_max_packet_size) {
  uint32_t val, dir;
  uint8_t num;

  num  = ep_addr & ARM_USB_ENDPOINT_NUMBER_MASK;
  val  = ep_max_packet_size & ARM_USB_ENDPOINT_MAX_PACKET_SIZE_MASK;

  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) dir = 16;
  else                                           dir = 0;

  if (num == 0) return ARM_USBD_OK;

  // Check if endpoint is enabled 
  if ((RTE_USB_DEV_EP & (1 << num)) == 0) return ARM_USBD_ERROR;

  // Check if endpoint supports requested type
  switch (ep_type) {
    case ARM_USB_ENDPOINT_ISOCHRONOUS:
      if ((RTE_USB_DEV_EP_ISO  & (1 << (num + dir))) == 0)
        return ARM_USBD_ERROR;
      break;
    case ARM_USB_ENDPOINT_BULK:
      if ((RTE_USB_DEV_EP_BULK & (1 << (num + dir))) == 0)
        return ARM_USBD_ERROR;
      break;
    case ARM_USB_ENDPOINT_INTERRUPT:
      if ((RTE_USB_DEV_EP_INT  & (1 << (num + dir))) == 0)
        return ARM_USBD_ERROR;
      break;
    default:
      break;
  }

  // IN Endpoint Configuration
  if (dir) {

    // Set IN Endpoint resources
    InEndpoint[num].buffer        = NULL;
    InEndpoint[num].dataSize      = 0;
    InEndpoint[num].maxPacketSize = val;
    InEndpoint[num].startRequired = false;
#if (USB_INT_IN_EP != 0)
    if (ep_type == ARM_USB_ENDPOINT_INTERRUPT) {
      InEndpoint[num].startRequired = true;
    }
#endif
    InEndpointNAK  &= ~(1 << num);

    // Configure IN endpoint
    USB_DIEPCTL(num) = (num     <<  _USB_DIEP_CTL_TXFNUM_SHIFT) | // Fifo Number
                       (ep_type <<  _USB_DIEP_CTL_EPTYPE_SHIFT) | // Endpoint Type
                       (val);                                     // Max Packet Size

    // Set DATA0 PID for Interrupt or Bulk Endpoint
    if (ep_type >= ARM_USB_ENDPOINT_BULK) {
      USB_DIEPCTL(num) |= USB_DIEP_CTL_SETD0PIDEF;
    }

    USB_DIEPCTL(num)   |= USB_DIEP_CTL_USBACTEP;  // Activate Endpoint

    if (USB_DIEPCTL(num) & USB_DIEP_CTL_EPENA) {
      USB_DIEPCTL(num) |= USB_DIEP_CTL_EPDIS;     // Disable endpoint
    }
    USB_DIEPCTL(num)   |= USB_DIEP_CTL_SNAK;      // Set Endpoint NAK

    USBD_FlushInEpFifo (num | ARM_USB_ENDPOINT_DIRECTION_MASK);

//     // Isochronous IN Endpoint Configuration
//     if (USB_EP_IN_TYPE(num) == USB_ENDPOINT_ISOCHRONOUS) {
//       USB->GINTMSK |= USB_GINTMSK_INCOMPISOINMSK;  // Enable IISOIXFR

//       // Regarding FrameNumber, set Frame
//       if (GetFrameNumber() & 1) USB_DIEPCTL(num) |= USB_DIEP_CTL_SETD0PIDEF;
//       else                      USB_DIEPCTL(num) |= USB_DIEP_CTL_SETD1PIDOF;

//       // Enable Endpoint and Clear NAK
//       USB_DIEPCTL(num) |= USB_DIEP_CTL_EPENA | USB_DIEP_CTL_CNAK;
//     }

    // Enable IN endpoint interrupt
    USB->DAINTMSK |= (1 << num) & 0x0000FFFF;


  // OUT Endpoint Configuration
  } else {
    // Set OUT Endpoint resources
    OutEndpoint[num].buffer         = NULL;
    OutEndpoint[num].dataSize       = 0;
    OutEndpoint[num].maxPacketSize  = val;

    // Configure OUT endpoint
    USB_DOEPCTL(num) = (ep_type <<  _USB_DOEP_CTL_EPTYPE_SHIFT)|  // Endpoint Type
                        USB_DOEP_CTL_SNAK                      |  // Set NAK
                        val;                                      // Max Packet Size

    // Set DATA0 PID for Interrupt or Bulk Endpoint
    if (ep_type >= ARM_USB_ENDPOINT_BULK) {
      USB_DOEPCTL(num) |= USB_DOEP_CTL_SETD0PIDEF;
    }

//     // Isochronous OUT Endpoint Configuration
//     if (USB_EP_OUT_TYPE(num) == USB_ENDPOINT_ISOCHRONOUS) {
//       USB->GINTMSK |= (1 << 15);          // Enable End of Periodic Frame Interrupt
//     }

    USB_DOEPCTL(num) |= USB_DOEP_CTL_USBACTEP;      // Activate Endpoint

   // Enable OUT endpoint interrupt
   USB->DAINTMSK     |= (1 << (num + 16)) & 0xFFFF0000;
  }

  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS EndpointUnconfigure (uint8_t ep_addr)
  \brief       Unconfigure USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS EndpointUnconfigure (uint8_t ep_addr) {
  uint32_t /*num, IsoEpEnCnt,*/ tick;

//  IsoEpEnCnt = 0;

  // Unconfigure IN Endpoint
  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
    ep_addr &= ARM_USB_ENDPOINT_NUMBER_MASK;

    USB->DAINTMSK &= ~(1 << ep_addr);   // Disable IN EP interrupt
//     // Count Active Isochronous IN Endpoints
//     if (USB_EP_IN_TYPE(ep_addr) == USB_ENDPOINT_ISOCHRONOUS) {
//       for (num = 1; num <= USBD_EP_NUM; num++) {
//         if (USB_DIEPCTL(num) & USB_DIEP_CTL_USBACTEP) {
//           if (USB_EP_IN_TYPE(num) == USB_ENDPOINT_ISOCHRONOUS) {
//             IsoEpEnCnt++;
//           }
//         }
//       }
//       // If Last Active Isochronous IN Endpoint, Disable INCOMPISOIN
//       if (IsoEpEnCnt == 1) USB->GINTMSK &= ~USB_GINTMSK_INCOMPISOINMSK;
//     }

    if (USB_DIEPCTL(ep_addr) & USB_DIEP_CTL_EPENA)
      USB_DIEPCTL(ep_addr)  |=  USB_DIEP_CTL_EPDIS;       // Disable Endpoint
    USB_DIEPCTL(ep_addr)    |=  USB_DIEP_CTL_SNAK;        // Set Endpoint NAK

    if (ep_addr) 
      USB_DIEPCTL(ep_addr)  &= ~USB_DIEP_CTL_USBACTEP;    // Deactivate Endpoint

    // Reset IN endpoint resources
    InEndpoint[ep_addr].buffer        = NULL;
    InEndpoint[ep_addr].dataSize      = 0;
    InEndpoint[ep_addr].maxPacketSize = 0;
    InEndpoint[ep_addr].startRequired = false;
    InEndpointNAK &= ~(1 << ep_addr);

  // Unconfigure OUT Endpoint
  } else {
    USB->DAINTMSK &= ~(1 << (ep_addr + 16));    // Disable IN EP interrupt

//     // Count Active Isochronous OUT Endpoints
//     if (USB_EP_OUT_TYPE(ep_addr) == USB_ENDPOINT_ISOCHRONOUS) {
//       for (num = 1; num <= USBD_EP_NUM; num++) {
//         if (USB_DOEPCTL(num) & USB_DOEP_CTL_USBACTEP) {
//           if (USB_EP_OUT_TYPE(num) == USB_ENDPOINT_ISOCHRONOUS) {
//             IsoEpEnCnt++;
//           }
//         }
//       }
//       // If Last Active Isochronous OUT Endpoint, Disable EOPF
//       if (IsoEpEnCnt == 1) USB->GINTMSK &= ~(1 << 15);
//     }

    // Set Global Out Nak
    USB->DCTL |= USB_DCTL_SGOUTNAK;
    tick = osKernelSysTick();
    do {
      if (USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF) break;
    } while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(10000));

    if ((USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF) == 0) return ARM_USBD_ERROR;

    USB_DOEPCTL(ep_addr) |= USB_DOEP_CTL_SNAK;        // Set Endpoint NAK

    if (ep_addr) {
      // Disable OUT endpoint
      if (USB_DOEPCTL(ep_addr) & USB_DOEP_CTL_EPENA) {// If Endpoint is Enabled
        USB_DOEPCTL(ep_addr)  |=  USB_DOEP_CTL_EPDIS; // Disable Endpoint

        tick = osKernelSysTick();
        do {
          if (USB_DOEPINT(ep_addr) & USB_DOEP_INT_EPDISBLD) break;
        } while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(10000));

        if ((USB_DOEPINT(ep_addr) & USB_DOEP_INT_EPDISBLD) == 0) return ARM_USBD_ERROR;
      }

      USB_DOEPCTL(ep_addr) &= ~USB_DOEP_CTL_USBACTEP; // Deactivate Endpoint
    }

    USB->DCTL |= USB_DCTL_CGOUTNAK;                   // Clear Global OUT NAK

    // Reset OUT endpoint resources
    OutEndpoint[ep_addr].buffer        = NULL;
    OutEndpoint[ep_addr].dataSize      = 0;
    OutEndpoint[ep_addr].maxPacketSize = 0;
  }
  return ARM_USBD_OK;
}

/**
  \fn          ARM_USBD_STATUS EndpointStall (uint8_t ep_addr, bool stall)
  \brief       Set/Clear Stall for USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   stall specifies operation
                - \b false Clear
                - \b true Set
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS EndpointStall (uint8_t ep_addr, bool stall) {
  uint32_t tick;

  if (stall) {
    // IN Endpoint stall
    if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
      ep_addr &= ARM_USB_ENDPOINT_NUMBER_MASK;

      if (USB_DIEPCTL(ep_addr) & USB_DIEP_CTL_EPENA)
        USB_DIEPCTL(ep_addr)  |= USB_DIEP_CTL_STALL | USB_DIEP_CTL_EPDIS;
      else 
        USB_DIEPCTL(ep_addr)  |= USB_DIEP_CTL_STALL;

      USBD_FlushInEpFifo (ep_addr | ARM_USB_ENDPOINT_DIRECTION_MASK);

    // OUT Endpoint Stall
    } else {
      USB->DCTL |= USB_DCTL_SGOUTNAK;               // Set Global OUT NAK
      tick = osKernelSysTick();
      do {
        if (USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF) break;
      } while ((osKernelSysTick() - tick) < osKernelSysTickMicroSec(10000));

      if ((USB->GINTSTS & USB_GINTSTS_GOUTNAKEFF) == 0) return ARM_USBD_ERROR;

      // Stall Out endpoint
      if (USB_DOEPCTL(ep_addr) & USB_DOEP_CTL_EPENA)
        USB_DOEPCTL(ep_addr)  |= USB_DOEP_CTL_STALL | USB_DOEP_CTL_EPDIS;
      else
        USB_DOEPCTL(ep_addr)  |= USB_DOEP_CTL_STALL;

      USB->DCTL |= USB_DCTL_CGOUTNAK;               // Clear global NAK
    }
  } else {
    // Clear IN Endpoint stall
    if (ep_addr & 0x80) {
      ep_addr &= ~0x80;

      if (USB_DIEPCTL(ep_addr) &  USB_DIEP_CTL_EPENA) // If Endpoint enabled
        USB_DIEPCTL(ep_addr)   |= USB_DIEP_CTL_EPDIS; // Disable Endpoint

      USBD_FlushInEpFifo (ep_addr | ARM_USB_ENDPOINT_DIRECTION_MASK);

      // Set DATA0 pid for interrupt and bulk endpoint
      if (((USB_DIEPCTL(ep_addr) & _USB_DIEP_CTL_EPTYPE_MASK) >> _USB_DIEP_CTL_EPTYPE_SHIFT) > 1)
        USB_DIEPCTL(ep_addr) |= USB_DIEP_CTL_SETD0PIDEF;

      USB_DIEPCTL(ep_addr) &= ~USB_DIEP_CTL_STALL;    // Clear Stall
 
    // Clear OUT Endpoint stall
    } else {
      // Set DATA0 pid for interrupt and bulk endpoint
      if (((USB_DOEPCTL(ep_addr) & _USB_DOEP_CTL_EPTYPE_MASK) >> _USB_DOEP_CTL_EPTYPE_SHIFT) > 1) {
        USB_DOEPCTL(ep_addr) |= USB_DOEP_CTL_SETD0PIDEF;
      }
      USB_DOEPCTL(ep_addr) &= ~USB_DOEP_CTL_STALL;    // Clear stall
    }
  }

  return ARM_USBD_OK;
}

/**
\fn          void USBD_HW_EndpointReadSet (uint8_t ep_addr, uint8_t *buf, uint32_t len)
\brief       Set Endpoint for next read.
\param[in]   ep_addr specifies Endpoint Address
              ep_addr.0..3: Address
              ep_addr.7:    Direction
*/
static void USBD_HW_EndpointReadSet (uint8_t ep_addr) {
  uint32_t sz;
  
  // Set packet count and transfer size
  if  (OutEndpoint[ep_addr].dataSize > OutEndpoint[ep_addr].maxPacketSize) sz = OutEndpoint[ep_addr].maxPacketSize;
  else                                                                     sz = OutEndpoint[ep_addr].dataSize;
  if (ep_addr != 0) {
    USB_DOEPTSIZ(ep_addr) = (1 << _USB_DOEP_TSIZ_PKTCNT_SHIFT) | 
                             sz;
  } else {
    USB_DOEPTSIZ(ep_addr) = (1 << _USB_DOEP0TSIZ_PKTCNT_SHIFT) | 
                            (3 << _USB_DOEP0TSIZ_SUPCNT_SHIFT) |
                             sz;
  }

//   // Set correct frame for isochronous endpoint
//   if (USB_EP_OUT_TYPE(ep_addr) == USB_ENDPOINT_ISOCHRONOUS) {
//     if ((GetFrameNumber() & 1)) USB_DOEPCTL(ep_addr) |= USB_DOEP_CTL_SETD0PIDEF;
//     else                        USB_DOEPCTL(ep_addr) |= USB_DOEP_CTL_SETD1PIDOF;
//   }

  // Clear NAK  and enable endpoint
  USB_DOEPCTL(ep_addr) |= USB_DOEP_CTL_EPENA | USB_DOEP_CTL_CNAK;
}

/**
\fn          ARM_USBD_STATUS EndpointReadStart (uint8_t ep_addr, uint8_t *buf, uint32_t len)
\brief       Start USB Endpoint Read operation.
\param[in]   ep_addr specifies Endpoint Address
              ep_addr.0..3: Address
              ep_addr.7:    Direction
\param[out]  buf specifies buffer for data read from Endpoint
\param[in]   len specifies buffer length
\return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS EndpointReadStart (uint8_t ep_addr, uint8_t *buf, uint32_t len) {

  if (OutEndpoint[ep_addr].maxPacketSize == 0)
    return ARM_USBD_ERROR;
  if (OutEndpoint[ep_addr].bufferIndex != 0)
    return ARM_USBD_ERROR;

  OutEndpoint[ep_addr].bufferIndex = 0;
  OutEndpoint[ep_addr].buffer      = buf;
  OutEndpoint[ep_addr].dataSize    = len;

  USBD_HW_EndpointReadSet(ep_addr);

  return ARM_USBD_OK;
}

/**
  \fn          static int32_t ReadFromFifo (uint8_t ep_addr)
  \brief       Read data from USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   sz specifies data size to be read from fifo
  \return      number of data bytes read, error code if negative
*/
static int32_t ReadFromFifo (uint8_t ep_addr, uint32_t sz) {
  uint32_t i, residue;
  uint8_t  tmpBuf[4];

  // Check if Endpoint is activated and buffer available
  if ((USB_DOEPCTL(ep_addr) & USB_DOEP_CTL_USBACTEP) == 0) return 0;
  if (OutEndpoint[ep_addr].buffer == NULL)                 return 0;

  if (sz > OutEndpoint[ep_addr].dataSize)
    sz = OutEndpoint[ep_addr].dataSize;

//   // If Isochronous Endpoint
//   if (USB_EP_OUT_TYPE(ep_addr) == USB_ENDPOINT_ISOCHRONOUS) {
//   /* if data pid = data0 and number of packets in 
//      which this payload was received = 1, then data is valid */
//      if ((USB_DOEPTSIZ(ep_addr) & (USB_DOEPTSIZx_PKTCNT | USB_DOEPTSIZx_RXDPID)) != 0) sz = 0;
//   }

  // Copy data from fifo
  for (i = 0; i < (uint32_t)(sz / 4); i++) {
    *((__packed uint32_t *)(OutEndpoint[ep_addr].buffer + OutEndpoint[ep_addr].bufferIndex)) = *((uint32_t *)USB->FIFO0D);
    OutEndpoint[ep_addr].bufferIndex += 4;
  }
  // if data size is not equal n*4
  residue = sz % 4;
  if (residue != 0) {
    *((__packed uint32_t *)tmpBuf) = *((uint32_t *)USB->FIFO0D);
    for (i = 0; i < residue; i++) {
      OutEndpoint[ep_addr].buffer[OutEndpoint[ep_addr].bufferIndex++] = tmpBuf[i];
    }
  }

  if (sz != OutEndpoint[ep_addr].maxPacketSize) OutEndpoint[ep_addr].dataSize  = 0;
  else                                          OutEndpoint[ep_addr].dataSize -= sz;

  return (sz);
}

/**
  \fn          int32_t EndpointRead (uint8_t ep_addr, uint8_t *buf, uint32_t len)
  \brief       Read data from USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[out]  buf specifies buffer for data read from Endpoint
  \param[in]   len specifies buffer length
  \return      number of data bytes read, error code if negative
*/
static int32_t EndpointRead (uint8_t ep_addr, uint8_t *buf, uint32_t len) {
  uint32_t sz;

  sz = OutEndpoint[ep_addr].bufferIndex;
  OutEndpoint[ep_addr].buffer      = NULL;
  OutEndpoint[ep_addr].bufferIndex = 0;
  return (sz);
}

/**
  \fn          static int32_t WriteToFifo (uint8_t ep_addr)
  \brief       Write data to USB Endpoint fifo.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \return      number of data bytes written, error code if negative
*/
static int32_t WriteToFifo (uint8_t ep_addr) {
  uint32_t i;

  ep_addr &= ARM_USB_ENDPOINT_NUMBER_MASK;

  // Check if enough space in fifo
  if ((USB_DTXFSTS(ep_addr) * 4) < InEndpoint[ep_addr].dataSize) return 0;

  // Set transfer size and packet count
  USB_DIEPTSIZ(ep_addr) = (1 << _USB_DIEP_TSIZ_PKTCNT_SHIFT) |
                          (1 << _USB_DIEP_TSIZ_MC_SHIFT)     |
                           InEndpoint[ep_addr].dataSize;

//   // Set correct frame fo isochronous endpoint
//   if (USB_EP_IN_TYPE(ep_addr) == USB_ENDPOINT_ISOCHRONOUS) {
//     if (GetFrameNumber() & 1) USB_DIEPCTL(ep_addr) |= USB_DIEP_CTL_SETD0PIDEF;
//     else USB_DIEPCTL(ep_addr)                      |= USB_DIEP_CTL_SETD1PIDOF;
//   }

  // Enable endpoint and clear NAK
  USB_DIEPCTL(ep_addr) |= USB_DIEP_CTL_EPENA | USB_DIEP_CTL_CNAK;

  // Copy data to fifo
  for (i = 0; i < (uint32_t)((InEndpoint[ep_addr].dataSize+3)/4); i++) {
    USB_TX_FIFO(ep_addr) = *(__packed uint32_t *)InEndpoint[ep_addr].buffer;
    InEndpoint[ep_addr].buffer +=4;
  }

  InEndpoint[ep_addr].buffer = NULL;
  return InEndpoint[ep_addr].dataSize;
}

/**
  \fn          int32_t EndpointWrite (uint8_t ep_addr, const uint8_t *buf, uint32_t len)
  \brief       Write data to USB Endpoint.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \param[in]   buf specifies buffer with data to write to Endpoint
  \param[in]   len specifies buffer length
  \return      number of data bytes written, error code if negative
*/
int32_t EndpointWrite (uint8_t ep_addr, const uint8_t *buf, uint32_t len) {
  uint8_t num;

  num = ep_addr & 0x0F;

  // If endpoint is disabled, return 0
  if ((USB_DIEPCTL(num) & USB_DIEP_CTL_USBACTEP) == 0) return (0);

  if (len > InEndpoint[num].maxPacketSize)
    len = InEndpoint[num].maxPacketSize;

  InEndpoint[num].dataSize = len;

#if (USB_INT_IN_EP != 0)
  if (USB_EP_IN_TYPE(num) == ARM_USB_ENDPOINT_INTERRUPT) {
    // Check if buffer is empty
    if (InEndpoint[num].buffer != NULL) return (-ARM_USBD_ERROR);

    // Copy data to intermediate buffer
    memcpy((uint8_t *)(ptr_USB_IntInEpTmpBuf[num]), buf, len);
    InEndpoint[num].buffer = (uint8_t *)ptr_USB_IntInEpTmpBuf[num];

    if (InEndpoint[num].startRequired == true) {
      InEndpoint[num].startRequired = false;
      InEndpointNAK    |=  1 << num;
      USB_DIEPCTL(num) |=  USB_DIEP_CTL_SNAK;           // Set NAK
      USB->DIEPMSK     |=  USB_DIEPMSK_INEPNAKEFFMSK;   // Enable NAK effective interrupt
    }
  } else
#endif
  {
    InEndpoint[num].buffer = (uint8_t *)buf;
    WriteToFifo(ep_addr);
  }
  return (len);
}

/**
  \fn          ARM_USBD_STATUS EndpointAbort (uint8_t ep_addr)
  \brief       Abort current USB Endpoint transfer.
  \param[in]   ep_addr specifies Endpoint Address
                ep_addr.0..3: Address
                ep_addr.7:    Direction
  \return      \ref ARM_USBD_STATUS
*/
static ARM_USBD_STATUS EndpointAbort (uint8_t ep_addr) {
  if (ep_addr & ARM_USB_ENDPOINT_DIRECTION_MASK) {
    ep_addr &= ARM_USB_ENDPOINT_NUMBER_MASK;
    InEndpoint[ep_addr].buffer = NULL;
    if (USB_DIEPCTL(ep_addr) &  USB_DIEP_CTL_EPENA)   // If endpoint enabled
      USB_DIEPCTL(ep_addr)   |= USB_DIEP_CTL_EPDIS;   // disable endpoint
    USB_DIEPCTL(ep_addr)     |= USB_DIEP_CTL_SNAK;    // Set NAK

    USBD_FlushInEpFifo (ep_addr | ARM_USB_ENDPOINT_DIRECTION_MASK);

  } else {
    OutEndpoint[ep_addr].buffer = NULL;
    if (USB_DOEPCTL(ep_addr) &  USB_DOEP_CTL_EPENA)   // If endpoint enabled
      USB_DOEPCTL(ep_addr)   |= USB_DOEP_CTL_EPDIS;   // Disable endpoint
    USB_DOEPCTL(ep_addr)     |= USB_DOEP_CTL_SNAK;    // Set NAK
  }
  return ARM_USBD_OK;
}

/**
  \fn          uint16_t GetFrameNumber (void)
  \brief       Get current USB Frame Number.
  \return      Frame Number
*/
static uint16_t GetFrameNumber (void) {
  return ((USB->DSTS & _USB_DSTS_SOFFN_MASK) >> _USB_DSTS_SOFFN_SHIFT);
}

/**
  \fn          void USBD_IRQ (uint32_t gintsts)
  \brief       USB Device Interrupt Routine (IRQ).
*/
void USBD_IRQ (uint32_t gintsts) {
  uint32_t val, num, msk, sz;
//  static uint32_t IsoInIncomplete = 0;

  // Reset interrupt
  if (gintsts & USB_GINTSTS_USBRST) {
    UsbdState.active = false;
    USBD_Reset();
    cbDeviceEvent(ARM_USBD_EVENT_RESET);
    USB->GINTSTS = USB_GINTSTS_USBRST;
  }

  // Suspend interrupt
  if (gintsts & USB_GINTSTS_USBSUSP) {
    UsbdState.active = false;
    cbDeviceEvent(ARM_USBD_EVENT_SUSPEND);
    USB->GINTSTS = USB_GINTSTS_USBSUSP;
  }

  // Resume interrupt
  if (gintsts & USB_GINTSTS_WKUPINT) {
    UsbdState.active = true;
    cbDeviceEvent(ARM_USBD_EVENT_RESUME);
    USB->GINTSTS = USB_GINTSTS_WKUPINT;
  }

  // Speed enumeration completed
  if (gintsts & USB_GINTSTS_ENUMDONE) {
    UsbdState.speed     = ARM_USB_SPEED_FULL;
    UsbdState.connected = true;

    USB->DCTL    |= USB_DCTL_CGNPINNAK;      // Clear global IN NAK
    USB->DCTL    |= USB_DCTL_CGOUTNAK ;      // clear global OUT NAK
    USB->GINTSTS  = USB_GINTSTS_ENUMDONE;
  }

  if (gintsts & USB_GINTSTS_RXFLVL) {
    USB->GINTMSK &= ~USB_GINTSTS_RXFLVL;

    val =  USB->GRXSTSP;
    num =  val & 0x0F;
    sz  = (val >> 4) & 0x7FF;

    switch ((val >> 17) & 0x0F) {
      // Setup packet
      case 6:
        ReadFromFifo(0, sz);
      break;

      // OUT packet
      case 2:
        ReadFromFifo(num, sz);
        break;

      // Global OUT NAK
      case 1:
        break;

      // OUT transfer completed
      case 3:
        break;
      
      // SETUP transaction completed
      case 4:
        break;

      default:
        break;
    }
    USB->GINTMSK |= USB_GINTSTS_RXFLVL;
  }

  // OUT Packet
  if (gintsts & USB_GINTSTS_OEPINT) {
    msk = (((USB->DAINT & USB->DAINTMSK) >> 16) & 0xFFFF);
    num = 0;

    do {
      if ((msk >> num) & 1) {
        // Endpoint disabled
        if (USB_DOEPINT(num) & USB_DOEP_INT_EPDISBLD) {
//           if (USB_EP_OUT_TYPE(num) == USB_ENDPOINT_ISOCHRONOUS) {
//             // Set packet count and transfer size
//             USB_DOEPTSIZ(num) = (1 << _USB_DOEP_TSIZ_PKTCNT_SHIFT) |
//                                 (OutEndpoint[num].maxPacketSize);

//             // Set correct frame
//             if ((GetFrameNumber() & 1)) USB_DOEPCTL(num) |= USB_DOEP_CTL_SETD0PIDEF;
//             else                        USB_DOEPCTL(num) |= USB_DOEP_CTL_SETD1PIDOF;

//             USB_DOEPCTL(num) |= USB_DOEP_CTL_EPENA | USB_DOEP_CTL_CNAK;
//           }
          USB_DOEPINT(num) = USB_DOEP_INT_EPDISBLD;
        }

        // Setup phase done interrupt
        if (USB_DOEPINT(num) & USB_DOEP_INT_SETUP) {
          OutEndpoint[num].dataSize = 0;
          USB_DOEPINT(num) = USB_DOEP_INT_SETUP;
          cbEndpointEvent(num, ARM_USBD_EP_EVENT_SETUP);
        }

        // Transfer complete interrupt
        if (USB_DOEPINT(num) & USB_DOEP_INT_XFERCOMPL) {
          USB_DOEPINT(num) = USB_DOEP_INT_XFERCOMPL;
          if (OutEndpoint[num].dataSize != 0) {
            USBD_HW_EndpointReadSet(num);
          } else {
            if (USB_EP_OUT_TYPE(num) != ARM_USB_ENDPOINT_ISOCHRONOUS) {
              cbEndpointEvent(num, ARM_USBD_EP_EVENT_OUT);
            }
          }
        }
      }
      num++;
    } while (msk >> num);
  }

  // IN Packet
  if (gintsts & USB_GINTSTS_IEPINT) {
    msk = (USB->DAINT & USB->DAINTMSK & 0xFFFF);
    num = 0;

    do {
      if ((msk >> num) & 1) {
        // Endpoint Disabled
        if (USB_DIEPINT(num) & USB_DIEP_INT_EPDISBLD) {
          USB_DIEPINT(num) = USB_DIEP_INT_EPDISBLD;

//           if (USB_EP_IN_TYPE(num) == USB_ENDPOINT_ISOCHRONOUS) {
//             if ((IsoInIncomplete & (1 << num)) != 0) {
//               // Flush IN endpoint fifo and write write new data if available
//               USBD_FlushInEpFifo(num | USB_ENDPOINT_DIRECTION_MASK);
//               if (InEndpoint[num].buffer) WriteToFifo(num);
//               IsoInIncomplete &= ~(1 << num);
//             }
//           }
        }

        // IN endpoint NAK effective
#if (USB_INT_IN_EP)
       if (USB_DIEPINT(num) & USB_DIEP_INT_INEPNAKEFF) {
          if (USB_EP_IN_TYPE(num) == ARM_USB_ENDPOINT_INTERRUPT) {
            if (InEndpointNAK & (1 << num)) {
              InEndpointNAK &= ~ (1 << num);
              // if no more forced NAKs, disable IN NAK effective interrupt
              if (InEndpointNAK == 0) USB->DIEPMSK &= ~USB_DIEPMSK_INEPNAKEFFMSK;

              // If Data available, write Data
              if (InEndpoint[num].buffer) WriteToFifo(num);
            }
          }
          USB_DIEPINT(num) = USB_DIEP_INT_INEPNAKEFF;
        }
#endif

        // Transmit completed
        if (USB_DIEPINT(num) & USB_DIEP_INT_XFERCOMPL) {
          USB_DIEPINT(num) = USB_DIEP_INT_XFERCOMPL;

#if (USB_INT_IN_EP)
          if (USB_EP_IN_TYPE(num) == ARM_USB_ENDPOINT_INTERRUPT) {
            // If Data available, write Data
            if (InEndpoint[num].buffer) WriteToFifo(num);

            // If no data available next interrupt IN transfer must
            // synchronized on NAK
            else InEndpoint[num].startRequired = true;
          }
#endif
          cbEndpointEvent(num, ARM_USBD_EP_EVENT_IN);
        }
      }
      num++;
    } while (msk >> num);
  }
}


ARM_DRIVER_USBD Driver_USBD0 = {
  GetVersion,
  GetCapabilities,
  Initialize,
  Uninitialize,
  PowerControl,
  DeviceConnect,
  DeviceDisconnect,
  DeviceGetState,
  DeviceRemoteWakeup,
  DeviceSetAddress,
  DeviceConfigure,
  EndpointConfigure,
  EndpointUnconfigure,
  EndpointStall,
  EndpointReadStart,
  EndpointRead,
  EndpointWrite,
  EndpointAbort,
  GetFrameNumber
};
